# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Functions for logical sign computation.

find_logical_signs
logical_flip
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pecos.circuits import QuantumCircuit
    from pecos.simulators.sparsesim.state import SparseSim


def find_logical_signs(
    state: SparseSim,
    logical_circuit: QuantumCircuit,
    delogical_circuit: QuantumCircuit | None = None,
) -> int:
    """Find the sign of the logical operator.

    Args:
        state (SparseSim): The SparseSim state instance.
        logical_circuit (QuantumCircuit): The logical circuit to find the sign of.
        delogical_circuit (QuantumCircuit | None): Optional delogical circuit to check anti-commutation with.
    """
    if len(logical_circuit) != 1:
        msg = "Logical operators are expected to only have one tick."
        raise Exception(msg)

    stabs = state.stabs
    destabs = state.destabs

    logical_xs = set()
    logical_zs = set()
    logical_ys = set()

    for symbol, gate_locations, _ in logical_circuit.items():
        if symbol == "X":
            logical_xs.update(gate_locations)
        elif symbol == "Z":
            logical_zs.update(gate_locations)
        elif symbol == "Y":
            logical_xs.update(gate_locations)
            logical_zs.update(gate_locations)
            logical_ys.update(gate_locations)
        else:
            msg = f'Can not currently handle logical operator with operator "{symbol}"!'
            raise Exception(
                msg,
            )

    if delogical_circuit:  # Check the relationship between logical operator and delogical operator.
        if len(delogical_circuit) != 1:
            msg = "Delogical operators are expected to only have one tick."
            raise Exception(msg)

        delogical_xs = set()
        delogical_zs = set()

        for symbol, gate_locations, _ in delogical_circuit.items():
            if symbol == "X":
                delogical_xs.update(gate_locations)
            elif symbol == "Z":
                delogical_zs.update(gate_locations)
            elif symbol == "Y":
                delogical_xs.update(gate_locations)
                delogical_zs.update(gate_locations)
            else:
                msg = f'Can not currently handle logical operator with operator "{symbol}"!'
                raise Exception(
                    msg,
                )

        # Make sure the logical and delogical anti-commute

        anticom_x = len(logical_xs & delogical_zs) % 2  # Number of common elements modulo 2
        anticom_z = len(logical_zs & delogical_xs) % 2  # Number of common elements modulo 2

        if not ((anticom_x + anticom_z) % 2):
            print(f"logical Xs: {logical_xs} logical Zs: {logical_zs}")
            print(f"delogical Xs: {delogical_xs} delogical Zs: {delogical_zs}")
            msg = "Logical and delogical operators supplied do not anti-commute!"
            raise Exception(msg)

    # We want the supplied logical operator to be in the stabilizer group and
    #  the supplied delogical to not be in the stabilizers (we want it to end up being the logical op's destabilizer)

    # The following two function calls are wasteful because we will need some of what they discover... such as all the
    #  stabilizers that have destabilizers that anti-commute with the logical operator...
    #  But it is assumed that the user is not calling this function that often... so we can be wasteful...

    # Check logical is a stabilizer (we want to remove it from the stabilizers)

    # Find the anti-commuting destabilizers => stabilizers to give the logical operator
    # --------------------------
    build_stabs = set()

    for q in logical_xs:  # For qubits that have Xs in for the logical operator...
        build_stabs ^= destabs.col_z[q]  # Add in stabilizers that anti-commute for the logical operator's Xs

    for q in logical_zs:
        build_stabs ^= destabs.col_x[q]  # Add in stabilizers that anti-commute for the logical operator's Zs

    # If a stabilizer anticommutes an even number of times for the X and/or Z Paulis... it will not appear due to ^=

    # Confirm that the stabilizers chosen give the logical operator. If not... return with a failure = 1
    # --------------------------
    test_x = set()
    test_z = set()

    for stab in build_stabs:
        test_x ^= stabs.row_x[stab]
        test_z ^= stabs.row_z[stab]

    # Compare with logical operator
    test_x ^= logical_xs
    test_z ^= logical_zs

    if len(test_x) != 0 or len(test_z) != 0:
        # for stab in build_stabs:

        print(f"Logical op: xs - {logical_xs} and zs - {logical_zs}")
        msg = f"Failure due to not finding logical op! x... {test_x ^ logical_xs!s} z... {test_z ^ logical_zs!s}"
        raise Exception(msg)

    # Get the sign of the logical operator
    # --------------------------

    """
    num_minuses = len(anticom_destabs_col & stabs.signs_minus)
    num_is = len(anticom_destabs_col & stabs.signs_i)

    # Sign correction due ZX -> -XZ
    cumulative_x = set([])
    for row in anticom_destabs_col:
        num_minuses += len(stabs_row_z[row] & cumulative_x)

        # Update the row sum Paulis
        cumulative_x ^= stabs_row_x[row]

    if num_is % 4:  # Can only be 0 or 2
        num_minuses += 1

    meas_outcome = num_minuses % 2
    """

    num_logical_is = len(logical_ys) % 2

    num_minuses = len(build_stabs & stabs.signs_minus)
    num_is = len(build_stabs & stabs.signs_i) - num_logical_is

    # Sign correction due ZX -> -XZ
    cumulative_x = set()
    for row in build_stabs:
        num_minuses += len(stabs.row_z[row] & cumulative_x)
        cumulative_x ^= stabs.row_x[row]

    if num_is % 4:  # Can only be 0 or 2
        num_minuses += 1

    if num_is % 2:
        msg = "Final sign should not have an imaginary sign!"
        raise Exception(msg)

    logical_sign = num_minuses % 2

    """
    # First, get the minus signs in front of the stabilizers
    logical_minus = len(build_stabs & stabs.signs_minus)

    print('num stab minuses', logical_minus)

    # Second, see get the imaginary signs in front of the stabilizers
    logical_i = len(build_stabs & stabs.signs_i)

    print('num stab is', logical_i)

    # Translate the Ws to Ys... W = -i(iW) = -iY => For each Y add another -1 and +i.
    # logical_ws = logical_xs & logical_zs
    num_ys = 0
    for stab in build_stabs:
        num_ys += len(stabs.row_x[stab] & stabs.row_z[stab])

    # num_ys = len(logical_ws)  # number of -is acquired

    print('num -is', num_ys, 'xs')

    logical_minus += num_ys
    logical_i += num_ys

    print('num minuses', logical_minus)
    print('num is', logical_i)

    # Do (-1)^even = 1 -> 0, (-1)^odd = -1 -> 1
    logical_minus %= 2

    # Reinterpret number of is
    logical_i %= 4

    # num_is %4 = 0 => +1 => logical_i = 0, logical_minus += 0
    # num_is %4 = 1 => +i => logical_i = 1, logical_minus += 0

    if logical_i == 2:  # num_is %4 = 2 => -1 => logical_i = 0, logical_minus += 1
        logical_i = 0
        logical_minus += 1
    elif logical_i == 3:  # num_is %4 = 3 => -i => logical_i = 1, logical_minus += 1
        logical_i = 1
        logical_minus += 1

    if logical_i != 0:
        raise Exception('Logical operator has an imaginary sign... Not allowed if logical state is stabilized '
                        'by logical op!')
    """

    return logical_sign



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for SlrConverter Guppy functionality."""


from pecos.slr import CReg, Main, Parallel, QReg, SlrConverter
from pecos.slr.qeclib import qubit as qb
from pecos.slr.qeclib.steane.steane_class import Steane


def test_slr_converter_guppy_simple() -> None:
    """Test SlrConverter.guppy() with a simple program."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        qb.Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # Check that the generated code is valid Python
    # AST codegen uses simplified imports
    assert "from guppylang import guppy" in guppy_code
    assert "@guppy" in guppy_code
    # AST codegen uses array parameters
    assert "def main(q:" in guppy_code
    assert "quantum.h(" in guppy_code
    assert "quantum.cx(" in guppy_code


def test_slr_converter_guppy_does_not_have_undefined_variables() -> None:
    """Test that generated Guppy code doesn't contain undefined variables."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        qb.Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # The generated code should compile without undefined variable errors
    # We can't easily test actual compilation here, but we can check for common issues
    lines = guppy_code.split("\n")

    # Find all variable assignments and uses
    defined_vars = set()
    for line in lines:
        stripped_line = line.strip()
        if "=" in stripped_line and not stripped_line.startswith("#") and not stripped_line.startswith("@"):
            # Extract variable definitions (left side of =)
            left_side = stripped_line.split("=")[0].strip()
            vars_defined = [v.strip() for v in left_side.split(",")] if "," in left_side else [left_side]
            defined_vars.update(vars_defined)

    # Now check that we don't have obvious undefined variable usage
    # This is a basic check - the real test would be compilation
    for line in lines:
        # Skip comments and decorators
        if line.strip().startswith("#") or line.strip().startswith("@"):
            continue
        # Skip import and function definition lines
        if any(keyword in line for keyword in ["import", "from", "def ", "class ", "return"]):
            continue

    # At minimum, check that we don't reference common undefined variables
    undefined_vars = ["c_a", "c_a_0"]  # Common issues we've seen
    for var in undefined_vars:
        assert var not in guppy_code, f"Generated code contains undefined variable: {var}"


def test_slr_converter_hugr_simple() -> None:
    """Test SlrConverter.hugr() with a simple program."""
    prog = Main(
        q := QReg("q", 2),
        c := CReg("c", 2),
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        qb.Measure(q) > c,
    )

    # This should not raise an exception
    hugr = SlrConverter(prog).hugr()

    # Basic checks that we got a valid HUGR object
    assert hugr is not None
    # HUGR is now a Package object with modules
    assert hasattr(hugr, "modules")


def test_slr_converter_steane_guppy_generation() -> None:
    """Test that Steane code can generate Guppy code without undefined variables."""
    prog = Main(
        c := Steane("c"),
        c.px(),  # Simple Pauli-X operation
    )

    # This should generate valid Guppy code without undefined variables
    guppy_code = SlrConverter(prog).guppy()

    # AST codegen uses array parameters instead of local declarations
    # Check that c_a is declared as parameter or that c_a[i] is properly accessed
    # Either c_a: array[...] @owned in params, or c_a = array(...) in body
    c_a_in_params = "c_a: array[qubit" in guppy_code
    c_a_declared = "c_a =" in guppy_code or "c_a=" in guppy_code

    # If c_a appears in the code, it should be in params or declared
    assert ("c_a" not in guppy_code) or c_a_in_params or c_a_declared

    # Code should have quantum operations
    assert "quantum." in guppy_code


def test_slr_converter_steane_hugr_compilation() -> None:
    """Test that Steane code can compile to HUGR.

    This test verifies that the Steane code implementation can be successfully
    compiled to HUGR format through guppylang. The test ensures that:

    1. Ancilla arrays (like c_a) are properly detected and excluded from structs
    2. These arrays are passed to functions with @owned annotation
    3. Arrays are unpacked to individual variables to avoid MoveOutOfSubscriptError
    4. The unpacked variables are used instead of array indexing in function bodies

    The solution works by:
    - Detecting ancilla qubits based on usage patterns (frequent measurement/reset)
    - Excluding them from struct packing to keep them as separate arrays
    - Unpacking @owned ancilla arrays at the start of functions
    - Using the unpacked variables (e.g., c_a_0) instead of array access (c_a[0])

    Note: The guppy code generation itself works correctly, but the final
    compilation to HUGR fails due to API mismatch between guppylang-internals
    (expecting hugr.build module) and hugr 0.13.0 (which doesn't have it).
    """
    prog = Main(
        c := Steane("c"),
        c.px(),
    )

    # This should work once guppylang supports the required patterns
    hugr = SlrConverter(prog).hugr()
    assert hugr is not None
    assert hasattr(hugr, "modules")


def test_slr_converter_parallel_blocks_guppy() -> None:
    """Test Guppy generation with parallel blocks."""
    prog = Main(
        q := QReg("q", 4),
        c := CReg("c", 4),
        Parallel(
            qb.H(q[0]),
            qb.X(q[1]),
            qb.H(q[2]),
            qb.X(q[3]),
        ),
        qb.Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # Should contain the gates
    assert "quantum.h(" in guppy_code
    assert "quantum.x(" in guppy_code
    assert "quantum.measure" in guppy_code

    # Should not have undefined variables
    undefined_vars = ["c_a", "c_a_0"]
    for var in undefined_vars:
        assert var not in guppy_code, f"Generated code contains undefined variable: {var}"


def test_slr_converter_guppy_has_main_function() -> None:
    """Test that generated Guppy code has a proper main function."""
    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        qb.H(q[0]),
        qb.Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # Should have main function with array parameters
    assert "def main(" in guppy_code
    assert "@guppy" in guppy_code


def test_slr_converter_guppy_imports() -> None:
    """Test that generated Guppy code has correct imports."""
    prog = Main(
        q := QReg("q", 1),
        c := CReg("c", 1),
        qb.H(q[0]),
        qb.Measure(q) > c,
    )

    guppy_code = SlrConverter(prog).guppy()

    # AST codegen uses simplified imports
    required_imports = [
        "from guppylang import guppy",
        "from guppylang.std import quantum",
        "from guppylang.std.quantum import qubit",
    ]

    for imp in required_imports:
        assert imp in guppy_code, f"Missing import: {imp}"



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QASM regression tests for Steane non-flagged Z measurement."""

from collections.abc import Callable

from pecos.slr import CReg, QReg
from pecos.slr.qeclib.steane.meas.measure_z import NoFlagMeasureZ


def test_MeasureX(compare_qasm: Callable[..., None]) -> None:
    """Test Steane non-flagged Z measurement QASM regression."""
    q = QReg("q_test", 7)
    a = QReg("a_test", 1)
    out = CReg("out_test", 1)

    block = NoFlagMeasureZ(q, a, out)
    compare_qasm(block)



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""QASM regression tests for Steane Pauli state preparations."""

from collections.abc import Callable

from pecos.slr import CReg, QReg
from pecos.slr.qeclib.steane.preps.pauli_states import (
    LogZeroRot,
    PrepEncodingFTZero,
    PrepEncodingNonFTZero,
    PrepRUS,
    PrepZeroVerify,
)


def test_PrepEncodingNonFTZero(compare_qasm: Callable[..., None]) -> None:
    """Test Steane non-fault-tolerant zero state preparation QASM regression."""
    q = QReg("q_test", 7)
    block = PrepEncodingNonFTZero(q)
    compare_qasm(block)


def test_PrepZeroVerify(compare_qasm: Callable[..., None]) -> None:
    """Test Steane zero state verification QASM regression."""
    q = QReg("q_test", 7)
    a = QReg("a_test", 1)
    init_bit = CReg("init_bit", 1)
    for reset_ancilla in [True, False]:
        block = PrepZeroVerify(q, a[0], init_bit[0], reset_ancilla=reset_ancilla)
        compare_qasm(block, reset_ancilla)


def test_PrepEncodingFTZero(compare_qasm: Callable[..., None]) -> None:
    """Test Steane fault-tolerant zero state preparation QASM regression."""
    q = QReg("q_test", 7)
    a = QReg("a_test", 1)
    init_bit = CReg("init_bit_test", 1)

    for reset in [True, False]:
        block = PrepEncodingFTZero(q, a[0], init_bit[0], reset=reset)
        compare_qasm(block, reset)


def test_PrepRUS(compare_qasm: Callable[..., None]) -> None:
    """Test Steane repeat-until-success state preparation QASM regression."""
    q = QReg("q_test", 7)
    a = QReg("a_test", 1)
    init = CReg("init_test", 1)

    for limit in [1, 2, 3]:
        for state in ["-Z", "+Z", "+X", "-X", "+Y", "-Y"]:
            for first_round_reset in [True, False]:
                block = PrepRUS(
                    q,
                    a[0],
                    init[0],
                    limit,
                    state,
                    first_round_reset=first_round_reset,
                )
                compare_qasm(block, limit, state, first_round_reset)


def test_LogZeroRot(compare_qasm: Callable[..., None]) -> None:
    """Test Steane logical zero rotation QASM regression."""
    q = QReg("q_test", 7)

    for state in ["-Z", "+Z", "+X", "-X", "+Y", "-Y"]:
        block = LogZeroRot(q, state)
        compare_qasm(block, state)



"""Simple depolarizing error model implementation.

This module provides a simplified depolarizing error model for basic
quantum error correction simulations with uniform depolarizing noise
applied to quantum gates.
"""

# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

import pecos as pc
from pecos.error_models.noise_impl_old.gate_groups import one_qubits, two_qubits
from pecos.reps.pyphir.op_types import QOp

if TYPE_CHECKING:
    from collections.abc import Callable

    from pecos.protocols import MachineProtocol
    from pecos.reps.pyphir.block_types import SeqBlock

one_qubit_paulis = ["X", "Y", "Z"]

two_qubit_paulis = [
    "IX",
    "IY",
    "IZ",
    "XI",
    "XX",
    "XY",
    "XZ",
    "YI",
    "YX",
    "YY",
    "YZ",
    "ZI",
    "ZX",
    "ZY",
    "ZZ",
]


class SimpleDepolarizingErrorModel:
    """Parameterized error mode."""

    def __init__(self, error_params: dict) -> None:
        """Initialize a simple depolarizing error model.

        Args:
            error_params: Dictionary containing error parameters including:
                - p1: Single-qubit gate error probability
                - p2: Two-qubit gate error probability
                - p_meas: Measurement error probability
                - p_init: Initialization error probability
        """
        self.error_params = dict(error_params)
        self.machine = None
        self.num_qubits = None
        self._eparams = None

    def reset(self) -> SimpleDepolarizingErrorModel:
        """Reset error generator for another round of syndrome extraction."""
        return SimpleDepolarizingErrorModel(error_params=self.error_params)

    def init(self, num_qubits: int, machine: MachineProtocol | None = None) -> None:
        """Initialize the simple depolarizing error model.

        Args:
            num_qubits: Number of qubits in the system.
            machine: Optional machine protocol for hardware-specific behavior.
        """
        self.machine = machine
        self.num_qubits = num_qubits

        if not self.error_params:
            msg = "Error params not set!"
            raise Exception(msg)

        self._eparams = dict(self.error_params)
        self._scale()

    def _scale(self) -> None:
        # conversion from average error to total error
        self._eparams["p1"] *= 3 / 2
        self._eparams["p2"] *= 5 / 4

        if isinstance(self._eparams["p_meas"], tuple):
            self._eparams["p_meas"] = pc.mean(self._eparams["p_meas"])

    def shot_reinit(self) -> None:
        """Run all code needed at the beginning of each shot, e.g., resetting state."""

    def process(
        self,
        qops: list[QOp],
        _call_back: Callable[..., None] | None = None,
    ) -> list[QOp | SeqBlock]:
        """Process quantum operations and apply simple depolarizing errors.

        Args:
            qops: List of quantum operations to process.
            call_back: Optional callback function for additional processing.

        Returns:
            List of quantum operations with applied errors.
        """
        noisy_ops = []

        for op in qops:
            erroneous_ops = None

            # ########################################
            # INITS WITH X NOISE
            if op.name in {"init |0>", "Init", "Init +Z"}:
                erroneous_ops = [op]
                # Use fused operation to check and get error indices in one pass
                error_indices = pc.random.compare_indices(
                    len(op.args),
                    self._eparams["p_init"],
                )

                for idx in error_indices:
                    erroneous_ops.append(
                        QOp(name="X", args=[op.args[idx]], metadata={}),
                    )

            # ########################################
            # ONE QUBIT GATES
            if op.name in one_qubits:
                erroneous_ops = [op]
                # Use fused operation to check and get error indices in one pass
                error_indices = pc.random.compare_indices(
                    len(op.args),
                    self._eparams["p1"],
                )

                for idx in error_indices:
                    err = pc.random.choice(one_qubit_paulis, 1)[0]
                    erroneous_ops.append(
                        QOp(name=err[0], args=[op.args[idx]], metadata={}),
                    )

            # ########################################
            # TWO QUBIT GATES
            elif op.name in two_qubits:
                erroneous_ops = [op]
                # Use fused operation to check and get error indices in one pass
                error_indices = pc.random.compare_indices(
                    len(op.args),
                    self._eparams["p2"],
                )

                for idx in error_indices:
                    err = pc.random.choice(two_qubit_paulis, 1)[0]
                    loc1, loc2 = op.args[idx]
                    if err[0] != "I":
                        erroneous_ops.append(
                            QOp(name=err[0], args=[loc1], metadata={}),
                        )
                    if err[1] != "I":
                        erroneous_ops.append(
                            QOp(name=err[1], args=[loc2], metadata={}),
                        )

            # ########################################
            # MEASURE X NOISE
            elif op.name in {"measure Z", "Measure", "Measure +Z"}:
                erroneous_ops = []
                # Use fused operation to check and get error indices in one pass
                error_indices = pc.random.compare_indices(
                    len(op.args),
                    self._eparams["p_meas"],
                )

                for idx in error_indices:
                    erroneous_ops.append(
                        QOp(name="X", args=[op.args[idx]], metadata={}),
                    )

                erroneous_ops.append(op)

            else:
                msg = f"This error model doesn't handle gate: {op.name}!"
                raise Exception(msg)

            if erroneous_ops is None:
                noisy_ops.append(op)
            else:
                noisy_ops.extend(erroneous_ops)

        return noisy_ops



"""Tests for complex permutation scenarios in both QASM and QIR generation."""

import re

import pytest
from pecos.slr import CReg, If, Main, Permute, QReg, SlrConverter
from pecos.slr.qeclib import qubit

# QASM Tests


def test_complex_permutation_circuit() -> None:
    """Test a more complex circuit with multiple permutations at different stages."""
    prog = Main(
        a := QReg("a", 3),
        b := QReg("b", 3),
        c := QReg("c", 3),
        # Initial operations - Layer 1
        qubit.H(a[0]),  # Hadamard on a[0]
        qubit.X(b[1]),  # X gate on b[1]
        qubit.Z(c[2]),  # Z gate on c[2]
        # First permutation: rotate registers
        Permute(
            [a[0], a[1], a[2], b[0], b[1], b[2], c[0], c[1], c[2]],
            [b[0], b[1], b[2], c[0], c[1], c[2], a[0], a[1], a[2]],
        ),
        # Operations after first permutation - Layer 2
        # a[0] -> b[0], b[1] -> c[1], c[2] -> a[2]
        qubit.X(a[0]),  # X gate on a[0] -> should become X on b[0]
        qubit.Y(b[1]),  # Y gate on b[1] -> should become Y on c[1]
        qubit.Z(c[2]),  # Z gate on c[2] -> should become Z on a[2]
    )

    qasm = SlrConverter(prog).qasm()

    # Check that the permutation was applied correctly
    assert "h a[0];" in qasm.lower()  # Initial operation
    assert "x b[1];" in qasm.lower()  # Initial operation
    assert "z c[2];" in qasm.lower()  # Initial operation

    assert "x b[0];" in qasm.lower()  # After first permutation
    assert "y c[1];" in qasm.lower()  # After first permutation
    assert "z a[2];" in qasm.lower()  # After first permutation


def test_multiple_permutations_qasm() -> None:
    """Test multiple sequential permutations in QASM generation."""
    # Create a program with multiple sequential permutations
    a = QReg("a", 3)
    b = QReg("b", 3)

    prog = Main(
        a,
        b,
        # First permutation
        Permute(
            [a[0], a[1]],
            [a[1], a[0]],
        ),
        # Apply an operation
        qubit.H(a[0]),  # Should become H(a[1]) after first permutation
        # Second permutation
        Permute(
            [a[1], b[0]],
            [b[0], a[1]],
        ),
        # Apply another operation
        qubit.H(a[0]),  # Should still be H(a[1]) after first permutation only
        qubit.X(a[1]),  # Should become X(b[0]) after both permutations
    )

    qasm = SlrConverter(prog).qasm()

    # Print the QASM for debugging
    print("\nQASM Output:")
    print(qasm)

    # Verify that the QASM contains the correct permuted operations
    assert "h a[1];" in qasm  # First H gate
    # The second H gate is applied to a[0] which is mapped to b[0] after both permutations
    assert "h b[0];" in qasm  # Second H gate
    # The X gate is applied to a[1] which is mapped to a[0] after both permutations
    assert "x a[0];" in qasm  # X gate after both permutations


def test_permutation_with_conditional_qasm() -> None:
    """Test permutation with conditional operations in QASM generation."""
    # Create a program with permutation and conditional operations
    a = QReg("a", 2)
    b = CReg("b", 2)

    prog = Main(
        a,
        b,
        # Set a classical bit
        b[0].set(1),
        # Apply a permutation
        Permute(
            [a[0], a[1], b[0], b[1]],
            [a[1], a[0], b[1], b[0]],
        ),
        # Apply a conditional operation
        # After permutation: b[0] -> b[1], a[0] -> a[1]
        # So the condition should be on b[1] and the operation should be on a[1]
        If(b[0] == 1).Then(qubit.X(a[0])),
    )

    qasm = SlrConverter(prog).qasm()

    # Print the QASM for debugging
    print("\nQASM Output:")
    print(qasm)

    # Verify that the QASM contains the correct permuted operations
    assert "b[0] = 1;" in qasm  # The classical bit assignment happens before permutation
    # The condition and operation should both be permuted
    assert "if(b[1] == 1) x a[1];" in qasm


# QIR Tests


@pytest.mark.optional_dependency
def test_multiple_permutations_qir() -> None:
    """Test multiple sequential permutations in QIR generation."""
    # Create a program with multiple sequential permutations
    a = QReg("a", 3)
    b = QReg("b", 3)

    prog = Main(
        a,
        b,
        # First permutation
        Permute(
            [a[0], a[1]],
            [a[1], a[0]],
        ),
        # Apply an operation
        qubit.H(a[0]),  # Should become H(a[1]) after first permutation
        # Second permutation
        Permute(
            [a[1], b[0]],
            [b[0], a[1]],
        ),
        # Apply another operation
        qubit.H(a[0]),  # Should still be H(a[1]) after first permutation only
        qubit.X(a[1]),  # Should become X(b[0]) after both permutations
    )

    qir = SlrConverter(prog).qir()

    # Verify that the QIR contains comments about the permutations
    assert "Permutation: a[0] -> a[1], a[1] -> a[0]" in qir
    assert "Permutation: a[1] -> b[0], b[0] -> a[1]" in qir

    # Extract the quantum operations
    h_calls = re.findall(
        r"call void @__quantum__qis__h__body\(%Qubit\* inttoptr \(i64 (\d+) to %Qubit\*\)\)",
        qir,
    )
    x_calls = re.findall(
        r"call void @__quantum__qis__x__body\(%Qubit\* inttoptr \(i64 (\d+) to %Qubit\*\)\)",
        qir,
    )

    # We should have at least two H calls and one X call
    assert len(h_calls) >= 2, f"Expected at least 2 H gate calls, found {len(h_calls)}"
    assert len(x_calls) >= 1, f"Expected at least 1 X gate call, found {len(x_calls)}"


@pytest.mark.optional_dependency
def test_permutation_with_conditional_qir() -> None:
    """Test permutation with conditional operations in QIR generation."""
    # Create a program with permutation and conditional operations
    a = QReg("a", 2)
    b = CReg("b", 2)

    prog = Main(
        a,
        b,
        # Set a classical bit
        b[0].set(1),
        # Apply a permutation
        Permute(
            [a[0], a[1], b[0], b[1]],
            [a[1], a[0], b[1], b[0]],
        ),
        # Apply a conditional operation
        # After permutation: b[0] -> b[1], a[0] -> a[1]
        # So the condition should be on b[1] and the operation should be on a[1]
        If(b[0] == 1).Then(qubit.X(a[0])),
    )

    qir = SlrConverter(prog).qir()

    # Verify that the QIR contains a comment about the permutation
    assert "Permutation: a[0] -> a[1], a[1] -> a[0], b[0] -> b[1], b[1] -> b[0]" in qir

    # Extract the set_creg_bit call
    set_creg_calls = re.findall(
        r"call void @set_creg_bit\(i1\* %(\w+), i64 (\d+), i1 1\)",
        qir,
    )

    # We should have at least one set_creg_bit call
    assert len(set_creg_calls) >= 1, "No set_creg_bit call found"

    # Get the register and index
    reg_name, index = set_creg_calls[0]

    # Verify that the set_creg_bit call is setting b[0] (not permuted, as it happens before the permutation)
    assert reg_name == "b", f"set_creg_bit applied to register {reg_name}, expected b"
    assert index == "0", f"set_creg_bit applied to index {index}, expected 0"

    # Extract the conditional X operation
    # In QIR, conditionals use __quantum__rt__array_get_element_ptr_1d to access the condition
    # and then branch based on the condition
    # This is a simplified check that just verifies an X gate is called somewhere after the condition check
    x_calls = re.findall(
        r"call void @__quantum__qis__x__body\(%Qubit\* inttoptr \(i64 (\d+) to %Qubit\*\)\)",
        qir,
    )

    # We should have at least one X call
    assert len(x_calls) >= 1, "No X gate call found"



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Quantum state representation for Qulacs simulator.

This module provides quantum state representation and management for the Qulacs simulator, including state vector
storage and manipulation using a pure Rust backend for high performance and thread safety.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos_rslib import simulators as rslib_sim

import pecos as pc
from pecos.simulators.qulacs import bindings
from pecos.simulators.sim_class_types import StateVector

if TYPE_CHECKING:
    from pecos import Array


class Qulacs(StateVector):
    """Wrapper of Qulacs state vector simulator using pure Rust backend."""

    def __init__(self, num_qubits: int, *, seed: int | None = None) -> None:
        """Initializes the state vector.

        Args:
            num_qubits (int): Number of qubits being represented.
            seed (int, optional): Random seed for deterministic behavior.
        """
        if not isinstance(num_qubits, int):
            msg = "``num_qubits`` should be of type ``int``."
            raise TypeError(msg)

        super().__init__()

        self.bindings = bindings.gate_dict
        self.num_qubits = num_qubits
        self.qulacs_state = rslib_sim.Qulacs(num_qubits, seed=seed)

        self.reset()

    def reset(self) -> Qulacs:
        """Reset the quantum state for another run without reinitializing."""
        # Initialize state vector to |0>
        self.qulacs_state.reset()
        return self

    @property
    def vector(self) -> Array:
        """Get the quantum state vector from Qulacs.

        Returns:
            The state vector as a PECOS array with complex values.
        """
        # Convert from [(real, imag), ...] tuples to complex array
        complex_tuples = self.qulacs_state.vector
        return pc.array(
            [complex(real, imag) for real, imag in complex_tuples],
            dtype="complex",
        )



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for AST resource counter."""

import pytest
from pecos.slr import CReg, Main, QReg, Repeat
from pecos.slr.ast import GateKind, slr_to_ast
from pecos.slr.ast.analysis import ResourceCounter, count_resources
from pecos.slr.qeclib import qubit as qb


class TestResourceCounterBasic:
    """Basic resource counting tests."""

    def test_empty_program(self) -> None:
        """Empty program has no resources."""
        prog = Main()
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.qubit_count == 0
        assert result.classical_bit_count == 0
        assert result.total_gates == 0

    def test_qubits_counted(self) -> None:
        """Qubits in allocators are counted."""
        prog = Main(
            _q := QReg("q", 5),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.qubit_count == 5
        assert result.qubits_by_allocator["q"] == 5

    def test_classical_bits_counted(self) -> None:
        """Classical bits in registers are counted."""
        prog = Main(
            _q := QReg("q", 1),
            _c := CReg("c", 3),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.classical_bit_count == 3
        assert result.bits_by_register["c"] == 3

    def test_multiple_registers(self) -> None:
        """Multiple registers are counted separately."""
        prog = Main(
            _q1 := QReg("q1", 2),
            _q2 := QReg("q2", 3),
            _c1 := CReg("c1", 1),
            _c2 := CReg("c2", 2),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.qubit_count == 5
        assert result.qubits_by_allocator["q1"] == 2
        assert result.qubits_by_allocator["q2"] == 3
        assert result.classical_bit_count == 3
        assert result.bits_by_register["c1"] == 1
        assert result.bits_by_register["c2"] == 2


class TestResourceCounterGates:
    """Gate counting tests."""

    def test_single_qubit_gates(self) -> None:
        """Single-qubit gates are counted."""
        prog = Main(
            q := QReg("q", 1),
            qb.Prep(q[0]),
            qb.H(q[0]),
            qb.X(q[0]),
            qb.Z(q[0]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.total_gates == 3  # H, X, Z (Prep is not a gate)
        assert result.single_qubit_gates == 3
        assert result.two_qubit_gates == 0

    def test_two_qubit_gates(self) -> None:
        """Two-qubit gates are counted."""
        prog = Main(
            q := QReg("q", 2),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
            qb.CX(q[0], q[1]),
            qb.CZ(q[0], q[1]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.total_gates == 2
        assert result.single_qubit_gates == 0
        assert result.two_qubit_gates == 2

    def test_mixed_gates(self) -> None:
        """Mixed single and two-qubit gates are counted correctly."""
        prog = Main(
            q := QReg("q", 2),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
            qb.X(q[1]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.total_gates == 3
        assert result.single_qubit_gates == 2  # H, X
        assert result.two_qubit_gates == 1  # CX

    def test_gate_counts_by_type(self) -> None:
        """Gates are counted by type."""
        prog = Main(
            q := QReg("q", 2),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
            qb.H(q[0]),
            qb.H(q[1]),
            qb.CX(q[0], q[1]),
            qb.H(q[0]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.gate_counts[GateKind.H] == 3
        assert result.gate_counts[GateKind.CX] == 1


class TestResourceCounterOperations:
    """Measurement and preparation counting tests."""

    def test_measurements_counted(self) -> None:
        """Measurements are counted."""
        prog = Main(
            q := QReg("q", 2),
            c := CReg("c", 2),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
            qb.Measure(q[0]) > c[0],
            qb.Measure(q[1]) > c[1],
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.measurement_count == 2

    def test_preparations_counted(self) -> None:
        """Preparations are counted."""
        prog = Main(
            q := QReg("q", 2),
            qb.Prep(q[0]),
            qb.Prep(q[1]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.preparation_count == 2


class TestResourceCounterControlFlow:
    """Control flow resource counting tests."""

    def test_repeat_multiplies_resources(self) -> None:
        """Repeat loop multiplies gate counts."""
        prog = Main(
            q := QReg("q", 1),
            qb.Prep(q[0]),
            Repeat(cond=5).block(
                qb.H(q[0]),
                qb.X(q[0]),
            ),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.gate_counts[GateKind.H] == 5
        assert result.gate_counts[GateKind.X] == 5
        assert result.total_gates == 10


class TestResourceCounterQEC:
    """QEC pattern resource counting tests."""

    def test_syndrome_extraction_resources(self) -> None:
        """Syndrome extraction resources are counted correctly."""
        prog = Main(
            data := QReg("data", 2),
            ancilla := QReg("ancilla", 1),
            c := CReg("c", 1),
            qb.Prep(data[0]),
            qb.Prep(data[1]),
            qb.Prep(ancilla[0]),
            qb.CX(data[0], ancilla[0]),
            qb.CX(data[1], ancilla[0]),
            qb.Measure(ancilla[0]) > c[0],
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)

        assert result.qubit_count == 3
        assert result.classical_bit_count == 1
        assert result.two_qubit_gates == 2
        assert result.measurement_count == 1
        assert result.preparation_count == 3


class TestResourceCounterClass:
    """Tests for the ResourceCounter class."""

    def test_counter_reusable(self) -> None:
        """Counter can be reused for multiple programs."""
        counter = ResourceCounter()

        prog1 = Main(_q := QReg("q", 2))
        prog2 = Main(_q := QReg("q", 5))

        ast1 = slr_to_ast(prog1)
        ast2 = slr_to_ast(prog2)

        result1 = counter.count(ast1)
        result2 = counter.count(ast2)

        assert result1.qubit_count == 2
        assert result2.qubit_count == 5

    def test_result_string_representation(self) -> None:
        """ResourceCount has useful string representation."""
        prog = Main(
            q := QReg("q", 2),
            _c := CReg("c", 1),
            qb.Prep(q[0]),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
        )
        ast = slr_to_ast(prog)

        result = count_resources(ast)
        result_str = str(result)

        assert "Qubits: 2" in result_str
        assert "Classical bits: 1" in result_str
        assert "Total gates: 2" in result_str



"""Pauli eigenstate preparation for the Steane 7-qubit code.

This module provides implementations for preparing logical Pauli eigenstates (|0⟩, |1⟩, |+⟩, |-⟩ |+i⟩, |-i⟩) in the
Steane 7-qubit code using both fault-tolerant and non-fault-tolerant encoding methods.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from pecos.slr import Barrier, Bit, Block, Comment, If, QReg, Qubit, Repeat
from pecos.slr.misc import Return
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit import Prep
from pecos.slr.qeclib.steane.gates_sq import sqrt_paulis
from pecos.slr.qeclib.steane.gates_sq.hadamards import H
from pecos.slr.qeclib.steane.gates_sq.paulis import X, Z
from pecos.slr.types import Array, QubitType


class PrepEncodingNonFTZero(Block):
    """Represents the non-fault-tolerant encoding circuit for the Steane code.

    Returns:
        array[qubit, 7]: The encoded 7-qubit register.
    """

    # Declare return type: returns the encoded qubit register
    block_returns = (Array[QubitType, 7],)

    def __init__(self, q: QReg) -> None:
        """Initialize PrepEncodingNonFTZero block for non-fault-tolerant zero state preparation.

        Args:
            q: Quantum register containing exactly 7 qubits for the Steane code.

        Raises:
            Exception: If the register does not contain exactly 7 qubits.
        """
        if len(q.elems) != 7:
            msg = f"Size of register {len(q.elems)} != 7"
            raise Exception(msg)

        super().__init__(
            qubit.H(
                q[0],
                q[4],
                q[6],
            ),
            Comment(),
            qubit.CX(
                (q[4], q[5]),
                (q[0], q[1]),
                (q[6], q[3]),
                (q[4], q[2]),
                (q[6], q[5]),
                (q[0], q[3]),
                (q[4], q[1]),
                (q[3], q[2]),
            ),
            # Explicitly declare return value
            Return(q),
        )


class PrepZeroVerify(Block):
    """Verify the initialization of InitEncodingNonFTZero."""

    def __init__(
        self,
        qubits: QReg,
        ancilla: Qubit,
        init_bit: Bit,
        *,
        reset_ancilla: bool = True,
    ) -> None:
        """Initialize PrepZeroVerify block for verification of zero state preparation.

        Args:
            qubits: Data register containing the 7 qubits of the Steane code.
            ancilla: Ancilla qubit used for verification.
            init_bit: Bit to store the verification result.
            reset_ancilla: Whether to reset the ancilla qubit. Defaults to True.
        """
        q = qubits
        a = ancilla
        c = init_bit

        super().__init__(
            Comment(),
            Barrier(a, q[1], q[3], q[5]),
            Comment("verification step"),
        )

        if reset_ancilla:
            self.extend(
                Comment(),
                Prep(a),
            )

        self.extend(
            qubit.CX(
                (q[5], a),
                (q[1], a),
                (q[3], a),
            ),
            qubit.Measure(a) > c,
            Comment(""),
        )


class PrepEncodingFTZero(Block):
    """Represents the fault-tolerant encoding circuit for the Steane code.

    This block prepares a logical zero state with verification. It consumes one ancilla
    qubit (measured during verification) and returns the remaining qubits.

    Returns:
        tuple[array[qubit, 2], array[qubit, 7]]: The ancilla register (size reduced from 3 to 2)
            and the data register (size unchanged at 7).

    Args:
        data (QReg[7]): Data register with 7 qubits (the logical Steane code)
        ancilla (QReg[3]): Ancilla register with 3 qubits
        init_bit (Bit): Bit to store initialization result
        reset (bool): Whether to reset qubits before preparation
    """

    # Declare return type: returns ancilla[2] and data[7]
    block_returns = (Array[QubitType, 2], Array[QubitType, 7])

    def __init__(
        self,
        data: QReg,
        ancilla: Qubit,
        init_bit: Bit,
        *,
        reset: bool = True,
    ) -> None:
        """Initialize PrepEncodingFTZero block for fault-tolerant zero state preparation.

        Args:
            data: Data register containing the 7 qubits of the Steane code.
            ancilla: Ancilla qubit used for verification.
            init_bit: Bit to store the initialization result.
            reset: Whether to reset qubits before preparation. Defaults to True.
        """
        q = data
        a = ancilla

        super().__init__()

        self.extend(
            Comment(),
            Barrier(q[0], q[1], q[2], q[3], q[4], q[5], q[6], a),
            Comment(),
        )

        if reset:
            self.extend(
                Prep(q),
                Prep(a),
                Barrier(q, a),
            )

        self.extend(
            PrepEncodingNonFTZero(data),
            # reset_ancilla to False because it is reset earlier
            PrepZeroVerify(data, ancilla, init_bit, reset_ancilla=False),
            # Explicitly declare return values (like Python's return statement)
            # Combined with block_returns annotation for robust type checking
            Return(a, q),
        )


class PrepRUS(Block):
    """Use repeat-until-success to initialize a logical qubit."""

    def __init__(
        self,
        q: QReg,
        a: Qubit,
        init: Bit,
        limit: int,
        state: str = "|0>",
        *,
        first_round_reset: bool = True,
    ) -> None:
        """Initialize PrepRUS block for repeat-until-success state preparation.

        Args:
            q: Data register containing the 7 qubits of the Steane code.
            a: Ancilla qubit used for verification.
            init: Bit to track initialization success.
            limit: Maximum number of preparation attempts.
            state: Target Pauli eigenstate to prepare. Defaults to '|0>'.
            first_round_reset: Whether to reset on first round. Defaults to True.
        """
        super().__init__(
            PrepEncodingFTZero(q, a, init, reset=first_round_reset),
            Repeat(limit - 1).block(
                If(init == 1).Then(
                    PrepEncodingFTZero(q, a, init, reset=True),
                ),
            ),
        )
        if limit == 1:
            self.extend(
                Comment(),
            )

        self.extend(
            # Rotate to the Pauli basis of choice
            LogZeroRot(q, state),
        )


class LogZeroRot(Block):
    """Rotate logical |0> to appropriate Pauli state.

    Returns:
        array[qubit, 7]: The rotated 7-qubit register in the target Pauli eigenstate.
    """

    # Declare return type: returns the rotated qubit register
    block_returns = (Array[QubitType, 7],)

    def __init__(self, q: QReg, state: str) -> None:
        """Initialize LogZeroRot block to rotate logical |0> to target Pauli state.

        Args:
            q: Data register containing the 7 qubits of the Steane code.
            state: Target state notation (e.g., '|0>', '|1>', '|+>', '|->', '|+i>', '|-i>').

        Raises:
            Exception: If the target state is not supported.
        """
        super().__init__()

        match state:
            case "|1>" | "-Z":
                self.extend(
                    X(q),
                )
            case "|0>" | "+Z" | "Z":
                pass
            case "|+>" | "+X" | "X":
                self.extend(
                    H(q),
                )
            case "|->" | "-X":
                self.extend(
                    H(q),
                    Z(q),
                )
            case "|+i>" | "+Y" | "Y":
                self.extend(
                    sqrt_paulis.SXdg(q),
                )
            case "|-i>" | "-Y":
                self.extend(
                    sqrt_paulis.SX(q),
                )
            case _:
                msg = f"Unsupported init state '{state}'"
                raise Exception(msg)

        # Explicitly declare return value
        self.extend(Return(q))



"""Demonstration of the rule-based unpacking decision tree.

This script shows how the new system provides clear, explainable decisions
about array unpacking, replacing the complex heuristic logic.
"""

from pecos.slr.gen_codes.guppy.ir_analyzer import ArrayAccessInfo
from pecos.slr.gen_codes.guppy.unpacking_rules import should_unpack_array


def demo_scenario(name: str, info: ArrayAccessInfo) -> None:
    """Demonstrate unpacking decision for a scenario."""
    print(f"\n{'='*70}")
    print(f"Scenario: {name}")
    print(f"{'='*70}")
    print(f"Array: {info.array_name} (size={info.size}, classical={info.is_classical})")
    print(f"Element accesses: {sorted(info.element_accesses)}")
    print(f"Elements consumed: {sorted(info.elements_consumed)}")
    print(f"Full array accesses: {info.full_array_accesses}")
    print(f"Has operations between: {info.has_operations_between}")
    print(f"Has conditionals between: {info.has_conditionals_between}")
    print()

    # Get decision with verbose output
    should_unpack_array(info, verbose=True)


def main() -> None:
    """Run demonstrations of various scenarios."""
    print("\n" + "=" * 70)
    print("RULE-BASED UNPACKING DECISION TREE DEMONSTRATION")
    print("=" * 70)

    # Scenario 1: Full array measurement
    info1 = ArrayAccessInfo(array_name="q", size=5, is_classical=False)
    info1.full_array_accesses.append(10)
    info1.element_accesses.add(0)  # Also has individual access
    demo_scenario("Full Array Measurement (prevents unpacking)", info1)

    # Scenario 2: Individual quantum measurements
    info2 = ArrayAccessInfo(array_name="q", size=3, is_classical=False)
    info2.element_accesses.add(0)
    info2.element_accesses.add(1)
    info2.elements_consumed.add(0)
    info2.elements_consumed.add(1)
    demo_scenario("Individual Quantum Measurements (requires unpacking)", info2)

    # Scenario 3: Operations after measurement
    info3 = ArrayAccessInfo(array_name="q", size=3, is_classical=False)
    info3.element_accesses.add(0)
    info3.elements_consumed.add(0)
    info3.has_operations_between = True
    demo_scenario("Operations After Measurement (requires unpacking)", info3)

    # Scenario 4: Conditional access
    info4 = ArrayAccessInfo(array_name="c", size=4, is_classical=True)
    info4.element_accesses.add(0)
    info4.element_accesses.add(1)
    info4.has_conditionals_between = True
    demo_scenario("Conditional Element Access (requires unpacking)", info4)

    # Scenario 5: Single element only
    info5 = ArrayAccessInfo(array_name="q", size=5, is_classical=False)
    info5.element_accesses.add(2)
    demo_scenario("Single Element Access (use direct indexing)", info5)

    # Scenario 6: Multiple classical accesses
    info6 = ArrayAccessInfo(array_name="c", size=4, is_classical=True)
    info6.element_accesses.add(0)
    info6.element_accesses.add(1)
    info6.element_accesses.add(2)
    demo_scenario("Multiple Classical Accesses (cleaner when unpacked)", info6)

    # Scenario 7: Partial array usage (high ratio)
    info7 = ArrayAccessInfo(array_name="q", size=5, is_classical=False)
    info7.element_accesses.add(0)
    info7.element_accesses.add(2)
    info7.element_accesses.add(4)
    demo_scenario("Partial Array Usage - High Ratio (60%)", info7)

    # Scenario 8: Partial array usage (low ratio)
    info8 = ArrayAccessInfo(array_name="q", size=10, is_classical=False)
    info8.element_accesses.add(0)
    info8.element_accesses.add(5)
    demo_scenario("Partial Array Usage - Low Ratio (20%)", info8)

    # Scenario 9: No individual access
    info9 = ArrayAccessInfo(array_name="q", size=5, is_classical=False)
    demo_scenario("No Individual Element Access", info9)

    print("\n" + "=" * 70)
    print("DEMONSTRATION COMPLETE")
    print("=" * 70)
    print("\nKey Improvements:")
    print("1. Clear, explainable decisions with reasoning")
    print("2. Explicit rules instead of complex heuristics")
    print("3. Easy to test and validate each rule")
    print("4. Maintainable and extensible")
    print()


if __name__ == "__main__":
    main()



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Parity check matrix generation for surface codes."""

import pecos
from pecos.qec.surface.layouts import (
    compute_x_stabilizer_supports,
    compute_z_stabilizer_supports,
)


def parity_matrix_x(d: int) -> pecos.Array:
    """Generate X stabilizer parity check matrix.

    Args:
        d: Code distance

    Returns:
        Binary matrix of shape (num_stab, num_data) where entry (i,j)=1
        if stabilizer i acts on qubit j
    """
    num_data = d * d
    stabs = compute_x_stabilizer_supports(d)
    num_stab = len(stabs)

    matrix = pecos.zeros((num_stab, num_data), dtype="int64")
    for stab in stabs:
        for q in stab.data_qubits:
            matrix[stab.index, q] = 1

    return matrix


def parity_matrix_z(d: int) -> pecos.Array:
    """Generate Z stabilizer parity check matrix.

    Args:
        d: Code distance

    Returns:
        Binary matrix of shape (num_stab, num_data)
    """
    num_data = d * d
    stabs = compute_z_stabilizer_supports(d)
    num_stab = len(stabs)

    matrix = pecos.zeros((num_stab, num_data), dtype="int64")
    for stab in stabs:
        for q in stab.data_qubits:
            matrix[stab.index, q] = 1

    return matrix



# Copyright 2018 The PECOS Developers
# Copyright 2018 National Technology & Engineering Solutions of Sandia, LLC (NTESS). Under the terms of Contract
# DE-NA0003525 with NTESS, the U.S. Government retains certain rights in this software.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Contains the default implementation of the logical instruction protocol."""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.protocols import LogicalInstructionProtocol
from pecos.qeccs.helper_functions import make_hashable_params
from pecos.qeccs.plot import plot_instr

if TYPE_CHECKING:
    from collections.abc import Iterator

    from pecos.circuits import LocationSet, QuantumCircuit
    from pecos.misc.symbol_library import JSONDict
    from pecos.protocols import QECCProtocol
    from pecos.typing import QECCInstrParams


class DefaultLogicalInstruction:
    """Default logical instruction class providing standard implementations.

    Logical instructions are circuits that implement specific quantum operations
    at the logical level. This class provides default implementations of the
    LogicalInstructionProtocol interface.

    Instruction implementations can inherit from this class to get the standard
    behavior, or implement the LogicalInstructionProtocol directly for custom behavior.
    """

    def __init__(
        self,
        qecc: QECCProtocol,
        symbol: str,
        **params: QECCInstrParams,
    ) -> None:
        """Initialize the LogicalInstruction with the given parameters.

        Args:
        ----
            qecc(QECC): The quantum error correcting code instance
            symbol(str): Symbol identifier for the instruction
            **params: Additional keyword arguments for instruction parameters
        """
        self.symbol = symbol
        self.qecc = qecc  # The QECC object this instruction belongs to.
        self.params = params
        self.gate_params = self.params  # Parameters used in defining the logical instruction.
        # TODO: should this be the same as the gate parameters?
        self.abstract_circuit = None  # Abstract representation of the circuit.
        self.circuit = None  # Compiled circuit.

        # The following assumes the role of ancilla and data qudits stays fixed during the instruction
        self.data_qudit_set = self.qecc.data_qudit_set  # set of qudit ids corresponding to data qudits.
        self.ancilla_qudit_set = self.qecc.ancilla_qudit_set  # set of qudit ids corresponding to ancilla qudits.
        # The ancilla set may differ from qecc. (might be a subset)

        self.params["data_qudit_set"] = self.data_qudit_set
        self.params["ancilla_qudit_set"] = self.ancilla_qudit_set

        # Logical operations
        # These are the expected initial and final logical operations
        self.initial_logical_ops = {}
        self.final_logical_ops = {}

        self.params_tuple = make_hashable_params(params)  # Used for hashing.

    def plot(self, **kwargs: object) -> None:
        """Creates a plot of the logical instruction.

        Returns: None

        """
        plot_instr(self, **kwargs)

    def _compile_circuit(
        self,
        abstract_circuit: QuantumCircuit,
        *args: object,
        **kwargs: object,
    ) -> None:
        """Create `circuit` instance from `abstract_circuit` instance for the logical instruction.

        If the instruction already has a `circuit` instance, do not bother compiling.
        """
        compiler = self.qecc.circuit_compiler
        compiler = self.params.get("circuit_compiler", compiler)

        self.circuit = compiler.compile(self, abstract_circuit, *args, **kwargs)

    def items(self) -> Iterator[tuple[str, LocationSet, JSONDict]]:
        """Yields: Yields the."""
        if self.circuit:
            return self.circuit.items()
        msg = ""
        raise Exception(msg)

    def __str__(self) -> str:
        """Return string representation of the logical instruction."""
        return f"[{self.qecc.name} {self.qecc.qecc_params}] - Logical instruction: '{self.symbol}' {self.params}"

    def __hash__(self) -> int:
        """Return hash value for use as dictionary key."""
        # The instruction is unique. A hash can be used to identify it.
        return hash(("instr", self.symbol, self.params_tuple))

    def __eq__(self, other: object) -> bool:
        """Check equality with another logical instruction."""
        # Check if other implements the LogicalInstructionProtocol
        if not isinstance(other, LogicalInstructionProtocol):
            return NotImplemented
        return (self.symbol, self.params_tuple, True) == (
            other.symbol,
            getattr(other, "params_tuple", None),
            hasattr(other, "circuit"),
        )

    def __ne__(self, other: object) -> bool:
        """Check inequality with another logical instruction."""
        return not (self == other)



"""Test suite for advanced quantum gates (Toffoli, CRz, etc.)."""

import pecos_rslib
import pytest
from guppylang import guppy
from guppylang.std.quantum import h, measure, pi, qubit

# Check if gates are available
try:
    from guppylang.std.quantum import crz, toffoli

    HAVE_ADVANCED_GATES = True
except ImportError:
    HAVE_ADVANCED_GATES = False

    # Define dummy functions for testing (never actually called - tests are skipped)
    # Type annotations match the actual guppylang function signatures
    def toffoli(q0: "qubit", q1: "qubit", q2: "qubit") -> None:  # type: ignore[name-defined]
        """Dummy toffoli gate for when advanced gates are not available."""

    def crz(q0: "qubit", q1: "qubit", angle: float) -> None:  # type: ignore[name-defined]
        """Dummy CRz gate for when advanced gates are not available."""


class TestThreeQubitGates:
    """Test three-qubit gates."""

    @pytest.mark.skipif(not HAVE_ADVANCED_GATES, reason="Advanced gates not available")
    def test_toffoli_gate(self) -> None:
        """Test Toffoli (CCX) gate."""

        @guppy
        def test_toffoli() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()
            h(q0)
            h(q1)
            toffoli(q0, q1, q2)
            return measure(q0), measure(q1), measure(q2)

        hugr = test_toffoli.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Toffoli should decompose into multiple gates
        assert "___rxy" in output
        assert "___rz" in output
        assert "___rzz" in output

        # Should have many operations (Toffoli needs many gates)
        ops_count = output.count("tail call void @___")
        assert ops_count >= 20, f"Toffoli should have many operations, got {ops_count}"


class TestControlledRotations:
    """Test controlled rotation gates."""

    @pytest.mark.skipif(not HAVE_ADVANCED_GATES, reason="Advanced gates not available")
    def test_crz_gate(self) -> None:
        """Test CRz gate with angle."""

        @guppy
        def test_crz() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            crz(q0, q1, pi / 4)
            return measure(q0), measure(q1)

        hugr = test_crz.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # CRz should use RZZ and RZ gates
        assert "___rzz" in output
        assert "___rz" in output


class TestCompilerFeatures:
    """Test compiler features and optimizations."""

    def test_transformation_passes_applied(self) -> None:
        """Test that transformation passes are applied (at least nominally)."""

        @guppy
        def simple() -> bool:
            q = qubit()
            h(q)
            return measure(q)

        hugr = simple.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should compile successfully
        assert "qmain" in output
        assert "___qalloc" in output

    def test_complex_circuit_compilation(self) -> None:
        """Test compilation of complex circuit with many gate types."""
        from guppylang.std.quantum import cx, cy, cz

        @guppy
        def complex_circuit() -> tuple[bool, bool, bool]:
            q0 = qubit()
            q1 = qubit()
            q2 = qubit()

            # Mix of gates
            h(q0)
            cx(q0, q1)
            cy(q1, q2)
            cz(q0, q2)

            # Measurements
            return measure(q0), measure(q1), measure(q2)

        hugr = complex_circuit.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should have all operation types
        assert "___rxy" in output
        assert "___rz" in output
        assert "___rzz" in output
        assert "___lazy_measure" in output
        assert "___qfree" in output

    def test_gate_count_optimization(self) -> None:
        """Verify that only used operations are declared."""
        from guppylang.std.quantum import cx

        @guppy
        def only_cnot() -> tuple[bool, bool]:
            q0 = qubit()
            q1 = qubit()
            h(q0)
            cx(q0, q1)
            return measure(q0), measure(q1)

        hugr = only_cnot.compile()
        output = pecos_rslib.compile_hugr_to_qis(hugr.to_bytes())

        # Should declare the operations we use
        assert "declare" in output
        assert "___rxy" in output  # For H and CX
        assert "___rz" in output  # For H and CX
        assert "___rzz" in output  # For CX

        # Count declarations vs actual usage
        declare_count = output.count("declare")
        # Should have reasonable number of declarations
        assert declare_count < 15, f"Too many declarations: {declare_count}"


# Test fallback for when advanced gates are not available
def test_advanced_gates_availability() -> None:
    """Check if advanced gates are available in guppylang."""
    import importlib.util

    # Check for Toffoli gate
    if importlib.util.find_spec("guppylang.std.quantum") is not None:
        try:
            from guppylang.std.quantum import toffoli

            assert True, "Toffoli gate is available"
        except (ImportError, AttributeError):
            pass  # Gate not available in this version

    # Check for CRz gate
    if importlib.util.find_spec("guppylang.std.quantum") is not None:
        try:
            from guppylang.std.quantum import crz

            assert True, "CRz gate is available"
        except (ImportError, AttributeError):
            pass  # Gate not available in this version



"""Steane 7-qubit quantum error correction code implementation.

This module provides the main Steane class that implements the Steane 7-qubit quantum error correction code, including
all necessary operations for fault-tolerant quantum computation.
"""

# Copyright 2024 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.slr import CReg, QReg, Vars
from pecos.slr.qeclib.steane.gates_sq import paulis, sqrt_paulis
from pecos.slr.qeclib.steane.gates_sq.hadamards import H
from pecos.slr.qeclib.steane.gates_tq import transversal_tq
from pecos.slr.qeclib.steane.meas.destructive_meas import Measure
from pecos.slr.qeclib.steane.preps.pauli_states import PrepRUS
from pecos.slr.qeclib.steane.syn_extract.bare import SynExtractBare
from pecos.slr.qeclib.steane.syn_extract.flagged import SynExtractFlagged

if TYPE_CHECKING:
    from pecos.slr import Bit, Block


class Steane(Vars):
    """A generic implementation of a Steane code and operations.

    This represents one particular choice of Steane protocols. For finer control construct your own class
    or utilize the library of Steane code protocols directly.
    """

    def __init__(
        self,
        name: str,
        default_rus_limit: int = 3,
        num_ancilla_qubits: int = 2,
    ) -> None:
        """Initialize a Steane code instance with associated quantum and classical registers.

        Args:
            name: Name prefix for all registers associated with this Steane code instance.
            default_rus_limit: Default limit for repeat-until-success procedures. Defaults to 3.
            num_ancilla_qubits: Number of ancilla qubits to allocate for syndrome extraction.

        Raises:
            ValueError: If provided ancilla register has fewer than 3 qubits.
        """
        super().__init__()
        self.check_indices = [[2, 1, 3, 0], [5, 2, 1, 4], [6, 5, 2, 3]]

        self.num_ancilla_qubits = num_ancilla_qubits

        self.d = QReg(f"{name}_d", 7)
        self.a = QReg(f"{name}_a", num_ancilla_qubits)

        self.verify_prep = CReg(f"{name}_verify_prep", 32)

        self.vars = [
            self.d,
            self.a,
        ]

        self.vars.extend(
            [
                self.verify_prep,
            ],
        )

        self.default_rus_limit = default_rus_limit

    def p(
        self,
        state: str,
        reject: Bit | None = None,
        rus_limit: int | None = None,
    ) -> Block:
        """Prepare a logical qubit in a logical Pauli basis state."""
        block = PrepRUS(
            q=self.d,
            a=self.a[0],
            init=self.verify_prep[0],
            limit=rus_limit or self.default_rus_limit,
            state=state,
            first_round_reset=True,
        )
        if reject is not None:
            block.extend(reject.set(self.verify_prep[0]))
        return block

    def x(self) -> Block:
        """Logical Pauli X gate."""
        return paulis.X(self.d)

    def y(self) -> Block:
        """Logical Pauli Y gate."""
        return paulis.Y(self.d)

    def z(self) -> Block:
        """Logical Pauli Z gate."""
        return paulis.Z(self.d)

    def h(self) -> Block:
        """Logical Hadamard gate."""
        return H(self.d)

    def sx(self) -> Block:
        """Sqrt of X."""
        return sqrt_paulis.SX(self.d)

    def sxdg(self) -> Block:
        """Adjoint of sqrt of X."""
        return sqrt_paulis.SXdg(self.d)

    def sy(self) -> Block:
        """Sqrt of Y."""
        return sqrt_paulis.SY(self.d)

    def sydg(self) -> Block:
        """Adjoint of sqrt of Y."""
        return sqrt_paulis.SYdg(self.d)

    def sz(self) -> Block:
        """Sqrt of Z. Also known as the S gate."""
        return sqrt_paulis.SZ(self.d)

    def szdg(self) -> Block:
        """Adjoint of Sqrt of Z. Also known as the Sdg gate."""
        return sqrt_paulis.SZdg(self.d)

    def cx(self, target: Steane) -> Block:
        """Logical CX."""
        return transversal_tq.CX(self.d, target.d)

    def cy(self, target: Steane) -> Block:
        """Logical CY."""
        return transversal_tq.CY(self.d, target.d)

    def cz(self, target: Steane) -> Block:
        """Logical CZ."""
        return transversal_tq.CZ(self.d, target.d)

    def m(
        self,
        meas_basis: str,
        meas: CReg,
        log: Bit,
        syn: CReg | None = None,
    ) -> Block:
        """Destructively measure the logical qubit in some Pauli basis."""
        block = Measure(
            q=self.d,
            meas_basis=meas_basis,
            log_raw=log,
            meas_creg=meas,
            barrier=False,
        )

        if syn is not None:
            block.extend(
                syn[0].set(meas[0] ^ meas[1] ^ meas[2] ^ meas[3]),
                syn[1].set(meas[1] ^ meas[2] ^ meas[4] ^ meas[5]),
                syn[2].set(meas[2] ^ meas[3] ^ meas[5] ^ meas[6]),
            )

        return block

    def syn_bare(self, syn: CReg) -> Block:
        """One single syndrome bit per check using bare syndrome extraction."""
        return SynExtractBare(self.d, self.a, self.check_indices, syn)

    def syn_flagged(
        self,
        syn_x: CReg,
        syn_z: CReg,
        flags_x: CReg,
        flags_z: CReg,
    ) -> Block:
        """One single syndrome bit and one single flag bit per check."""
        return SynExtractFlagged(
            self.d,
            self.a,
            self.check_indices,
            syn_x,
            syn_z,
            flags_x,
            flags_z,
        )



# Copyright 2023 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Block type definitions for PyPHIR intermediate representation.

This module defines block structures for PyPHIR (Python PECOS Medium-level Intermediate Representation) including
conditional blocks and control flow structures for quantum circuit execution.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from pecos.reps.pyphir.instr_type import Instr

if TYPE_CHECKING:
    from pecos.reps.pyphir.op_types import COp, Op, QOp


class Block(Instr):
    """General block type."""


class SeqBlock(Block):
    """A generic sequence block.

    This is not meant to indicate parallelism or lack of it but is just a way to structure operations/blocks.
    """

    def __init__(self, ops: list[Op | Block], metadata: dict | None = None) -> None:
        """Initialize a sequence block.

        Args:
            ops: List of operations or blocks to be executed in sequence.
            metadata: Optional metadata dictionary.
        """
        super().__init__(metadata=metadata)
        self.ops = ops


class QParallelBlock(SeqBlock):
    """A block to indicate that a collection of QOps are applied in parallel."""

    def __init__(self, ops: list[QOp], metadata: dict | None = None) -> None:
        """Initialize a parallel quantum operations block.

        Args:
            ops: List of quantum operations to be applied in parallel.
            metadata: Optional metadata dictionary.
        """
        super().__init__(ops=ops, metadata=metadata)


class IfBlock(Block):
    """If/else block."""

    def __init__(
        self,
        condition: COp,
        true_branch: list[Op],
        false_branch: list[Op] | None = None,
        metadata: dict | None = None,
    ) -> None:
        """Initialize an if/else block.

        Args:
            condition: Classical operation that evaluates to a boolean condition.
            true_branch: List of operations to execute if condition is true.
            false_branch: Optional list of operations to execute if condition is false.
            metadata: Optional metadata dictionary.
        """
        super().__init__(metadata=metadata)
        self.condition = condition
        self.true_branch = true_branch
        self.false_branch = false_branch



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Surface code geometry and parity matrices.

This module provides pure geometry computations for surface codes,
including stabilizer support calculation and parity check matrices.

Layouts:
    square_lattice: Standard square lattice (d^2 data qubits)
    rotated_lattice: Rotated lattice (more efficient)

Functions:
    compute_x_stabilizer_supports: Get X stabilizer qubit indices
    compute_z_stabilizer_supports: Get Z stabilizer qubit indices
    parity_matrix_x: Generate X parity check matrix
    parity_matrix_z: Generate Z parity check matrix
"""

from pecos.qec.surface.layouts import (
    StabilizerSupport,
    compute_rotated_x_stabilizers,
    compute_rotated_z_stabilizers,
    compute_x_stabilizer_supports,
    compute_z_stabilizer_supports,
    generate_nonrotated_surface_layout,
    generate_surface_layout,
    get_rotated_logical_x,
    get_rotated_logical_z,
)
from pecos.qec.surface.parity import (
    parity_matrix_x,
    parity_matrix_z,
)
from pecos.qec.surface.patch import (
    LogicalOperator,
    PatchGeometry,
    PatchOrientation,
    Stabilizer,
    SurfacePatch,
    SurfacePatchBuilder,
)

__all__ = [  # noqa: RUF022
    # Rotated lattice (most common, default)
    "compute_rotated_x_stabilizers",
    "compute_rotated_z_stabilizers",
    "generate_surface_layout",
    "get_rotated_logical_x",
    "get_rotated_logical_z",
    # Non-rotated lattice
    "StabilizerSupport",
    "compute_x_stabilizer_supports",
    "compute_z_stabilizer_supports",
    "generate_nonrotated_surface_layout",
    # Parity matrices
    "parity_matrix_x",
    "parity_matrix_z",
    # Patch classes
    "LogicalOperator",
    "PatchGeometry",
    "PatchOrientation",
    "Stabilizer",
    "SurfacePatch",
    "SurfacePatchBuilder",
]



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for AST pretty-printing."""

import pytest
from pecos.slr import CReg, If, Main, QAlloc, QReg, Repeat
from pecos.slr.ast import slr_to_ast
from pecos.slr.ast.nodes import (
    BinaryExpr,
    BinaryOp,
    BitRef,
    GateKind,
    GateOp,
    LiteralExpr,
    SlotRef,
    UnaryExpr,
    UnaryOp,
)
from pecos.slr.ast.pretty_print import (
    AstPrettyPrinter,
    format_expression,
    format_statement,
    pretty_print,
)
from pecos.slr.qeclib import qubit as qb


class TestPrettyPrintBasic:
    """Basic pretty-printing tests."""

    def test_empty_program(self) -> None:
        """Empty program prints correctly."""
        prog = Main()
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "Main(" in output
        assert output.endswith(")")

    def test_program_with_qreg(self) -> None:
        """Program with QReg declaration prints correctly."""
        prog = Main(
            _q := QReg("q", 2),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert 'q := QReg("q", 2)' in output

    def test_program_with_creg(self) -> None:
        """Program with CReg declaration prints correctly."""
        prog = Main(
            _q := QReg("q", 1),
            _c := CReg("c", 2),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert 'c := CReg("c", 2)' in output

    def test_indentation(self) -> None:
        """Output is properly indented."""
        prog = Main(
            q := QReg("q", 1),
            qb.H(q[0]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)
        lines = output.split("\n")

        # First line should not be indented
        assert lines[0] == "Main("
        # Content lines should be indented
        assert lines[1].startswith("    ")
        # Last line should not be indented
        assert lines[-1] == ")"

    def test_custom_indent(self) -> None:
        """Custom indent string is respected."""
        prog = Main(
            q := QReg("q", 1),
            qb.H(q[0]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast, indent="  ")  # 2 spaces
        lines = output.split("\n")

        assert lines[1].startswith("  ")
        assert not lines[1].startswith("    ")


class TestPrettyPrintGates:
    """Gate pretty-printing tests."""

    def test_single_qubit_gate(self) -> None:
        """Single-qubit gate prints correctly."""
        prog = Main(
            q := QReg("q", 1),
            qb.H(q[0]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "qb.H(q[0])" in output

    def test_pauli_gates(self) -> None:
        """Pauli gates print correctly."""
        prog = Main(
            q := QReg("q", 3),
            qb.X(q[0]),
            qb.Y(q[1]),
            qb.Z(q[2]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "qb.X(q[0])" in output
        assert "qb.Y(q[1])" in output
        assert "qb.Z(q[2])" in output

    def test_two_qubit_gate(self) -> None:
        """Two-qubit gate prints correctly."""
        prog = Main(
            q := QReg("q", 2),
            qb.CX(q[0], q[1]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "qb.CX(q[0], q[1])" in output

    def test_phase_gates(self) -> None:
        """Phase gates print correctly."""
        prog = Main(
            q := QReg("q", 2),
            qb.SZ(q[0]),  # S gate (sqrt Z)
            qb.T(q[1]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "qb.SZ(q[0])" in output
        assert "qb.T(q[1])" in output


class TestPrettyPrintMeasure:
    """Measurement pretty-printing tests."""

    def test_measure_without_result(self) -> None:
        """Measurement without result prints correctly."""
        gate = GateOp(
            gate=GateKind.H,
            targets=(SlotRef(allocator="q", index=0),),
        )

        output = format_statement(gate)

        assert "qb.H(q[0])" in output

    def test_measure_with_result(self) -> None:
        """Measurement with result prints correctly."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            qb.Measure(q[0]) > c[0],
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        # Check measurement is formatted
        assert "Measure" in output
        assert "q[0]" in output


class TestPrettyPrintControlFlow:
    """Control flow pretty-printing tests."""

    def test_if_statement(self) -> None:
        """If statement prints correctly."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1).Then(
                qb.X(q[0]),
            ),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "If(" in output
        assert ").Then(" in output
        assert "qb.X(q[0])" in output

    def test_if_else_statement(self) -> None:
        """If-else statement prints correctly."""
        prog = Main(
            q := QReg("q", 1),
            c := CReg("c", 1),
            If(c[0] == 1)
            .Then(
                qb.X(q[0]),
            )
            .Else(
                qb.Y(q[0]),
            ),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "If(" in output
        assert ").Then(" in output
        assert ").Else(" in output
        assert "qb.X(q[0])" in output
        assert "qb.Y(q[0])" in output

    def test_repeat_statement(self) -> None:
        """Repeat statement prints correctly."""
        prog = Main(
            q := QReg("q", 1),
            Repeat(cond=5).block(
                qb.H(q[0]),
            ),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "Repeat(cond=5).block(" in output
        assert "qb.H(q[0])" in output

    def test_nested_control_flow(self) -> None:
        """Nested control flow prints with proper indentation."""
        prog = Main(
            q := QReg("q", 2),
            c := CReg("c", 2),
            If(c[0] == 1).Then(
                If(c[1] == 1).Then(
                    qb.X(q[0]),
                ),
            ),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        # Count indentation levels
        lines = output.split("\n")
        max_indent = max(len(line) - len(line.lstrip()) for line in lines if line.strip())

        # Should have at least 3 levels of indentation for nested if
        assert max_indent >= 8  # 2 levels * 4 spaces


class TestPrettyPrintExpressions:
    """Expression pretty-printing tests."""

    def test_literal_int(self) -> None:
        """Integer literal formats correctly."""
        expr = LiteralExpr(value=42)

        output = format_expression(expr)

        assert output == "42"

    def test_literal_float(self) -> None:
        """Float literal formats correctly."""
        expr = LiteralExpr(value=3.14)

        output = format_expression(expr)

        assert "3.14" in output

    def test_literal_bool_true(self) -> None:
        """True literal formats correctly."""
        expr = LiteralExpr(value=True)

        output = format_expression(expr)

        assert output == "True"

    def test_literal_bool_false(self) -> None:
        """False literal formats correctly."""
        expr = LiteralExpr(value=False)

        output = format_expression(expr)

        assert output == "False"

    def test_binary_expr_eq(self) -> None:
        """Equality expression formats correctly."""
        expr = BinaryExpr(
            op=BinaryOp.EQ,
            left=LiteralExpr(value=1),
            right=LiteralExpr(value=2),
        )

        output = format_expression(expr)

        assert "==" in output
        assert "1" in output
        assert "2" in output

    def test_binary_expr_add(self) -> None:
        """Addition expression formats correctly."""
        expr = BinaryExpr(
            op=BinaryOp.ADD,
            left=LiteralExpr(value=3),
            right=LiteralExpr(value=4),
        )

        output = format_expression(expr)

        assert "+" in output

    def test_binary_expr_and(self) -> None:
        """And expression formats correctly."""
        expr = BinaryExpr(
            op=BinaryOp.AND,
            left=LiteralExpr(value=True),
            right=LiteralExpr(value=False),
        )

        output = format_expression(expr)

        assert "and" in output

    def test_unary_expr_not(self) -> None:
        """Not expression formats correctly."""
        expr = UnaryExpr(
            op=UnaryOp.NOT,
            operand=LiteralExpr(value=True),
        )

        output = format_expression(expr)

        assert "not" in output

    def test_unary_expr_neg(self) -> None:
        """Negation expression formats correctly."""
        expr = UnaryExpr(
            op=UnaryOp.NEG,
            operand=LiteralExpr(value=5),
        )

        output = format_expression(expr)

        assert "-" in output


class TestPrettyPrintReferences:
    """Reference pretty-printing tests."""

    def test_slot_ref(self) -> None:
        """SlotRef formats as allocator[index]."""
        slot = SlotRef(allocator="data", index=3)
        printer = AstPrettyPrinter()

        output = printer.visit_slot_ref(slot)

        assert output == "data[3]"

    def test_bit_ref(self) -> None:
        """BitRef formats as register[index]."""
        bit = BitRef(register="result", index=5)
        printer = AstPrettyPrinter()

        output = printer.visit_bit_ref(bit)

        assert output == "result[5]"


class TestPrettyPrintHierarchicalAllocators:
    """Hierarchical allocator pretty-printing tests."""

    def test_hierarchical_allocators(self) -> None:
        """Hierarchical allocators show parent relationship."""
        all_qubits = QAlloc(4, name="all")
        data = QAlloc(2, name="data", parent=all_qubits)
        ancilla = QAlloc(2, name="ancilla", parent=all_qubits)

        prog = Main(
            all_qubits,
            data,
            ancilla,
            qb.H(data[0]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        # Check parent relationship is shown
        assert "parent=" in output or "data" in output


class TestPrettyPrintQEC:
    """QEC pattern pretty-printing tests."""

    def test_syndrome_extraction(self) -> None:
        """Syndrome extraction pattern prints correctly."""
        prog = Main(
            data := QReg("data", 2),
            ancilla := QReg("ancilla", 1),
            c := CReg("c", 1),
            qb.CX(data[0], ancilla[0]),
            qb.CX(data[1], ancilla[0]),
            qb.Measure(ancilla[0]) > c[0],
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        # Check key elements
        assert 'data := QReg("data", 2)' in output
        assert 'ancilla := QReg("ancilla", 1)' in output
        assert "qb.CX(data[0], ancilla[0])" in output
        assert "qb.CX(data[1], ancilla[0])" in output

    def test_bell_state(self) -> None:
        """Bell state circuit prints correctly."""
        prog = Main(
            q := QReg("q", 2),
            qb.H(q[0]),
            qb.CX(q[0], q[1]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert "qb.H(q[0])" in output
        assert "qb.CX(q[0], q[1])" in output


class TestFormatStatement:
    """Tests for format_statement function."""

    def test_format_gate(self) -> None:
        """Single gate formats correctly."""
        gate = GateOp(
            gate=GateKind.H,
            targets=(SlotRef(allocator="q", index=0),),
        )

        output = format_statement(gate)

        assert output == "qb.H(q[0])"

    def test_format_two_qubit_gate(self) -> None:
        """Two-qubit gate formats correctly."""
        gate = GateOp(
            gate=GateKind.CZ,
            targets=(
                SlotRef(allocator="a", index=0),
                SlotRef(allocator="b", index=1),
            ),
        )

        output = format_statement(gate)

        assert output == "qb.CZ(a[0], b[1])"


class TestFormatExpression:
    """Tests for format_expression function."""

    def test_format_simple(self) -> None:
        """Simple expression formats correctly."""
        expr = LiteralExpr(value=100)

        output = format_expression(expr)

        assert output == "100"

    def test_format_complex(self) -> None:
        """Complex nested expression formats correctly."""
        expr = BinaryExpr(
            op=BinaryOp.ADD,
            left=BinaryExpr(
                op=BinaryOp.MUL,
                left=LiteralExpr(value=2),
                right=LiteralExpr(value=3),
            ),
            right=LiteralExpr(value=4),
        )

        output = format_expression(expr)

        # Should contain structure of expression
        assert "2" in output
        assert "3" in output
        assert "4" in output
        assert "*" in output
        assert "+" in output


class TestPrettyPrinterClass:
    """Tests for AstPrettyPrinter class."""

    def test_reusable(self) -> None:
        """Printer can be reused for multiple programs."""
        printer = AstPrettyPrinter()

        prog1 = Main(q := QReg("q", 1), qb.H(q[0]))
        prog2 = Main(r := QReg("r", 2), qb.X(r[0]))

        ast1 = slr_to_ast(prog1)
        ast2 = slr_to_ast(prog2)

        output1 = printer.print(ast1)
        output2 = printer.print(ast2)

        assert "q[0]" in output1
        assert "r[0]" in output2

    def test_indent_level_reset(self) -> None:
        """Indent level resets between prints."""
        printer = AstPrettyPrinter()

        prog = Main(
            q := QReg("q", 1),
            If(LiteralExpr(value=True)).Then(
                qb.H(q[0]),
            ),
        )
        ast = slr_to_ast(prog)

        # Print multiple times
        output1 = printer.print(ast)
        output2 = printer.print(ast)

        # Outputs should be identical
        assert output1 == output2


class TestEdgeCases:
    """Edge case tests."""

    def test_empty_if_body(self) -> None:
        """Test If with no operations in then body (unusual but valid)."""
        # This would require manual AST construction since SLR requires content
        # Skip - SLR requires at least one statement

    def test_multiple_allocators(self) -> None:
        """Multiple allocators are rendered correctly."""
        prog = Main(
            a := QReg("a", 2),
            b := QReg("b", 2),
            c := QReg("c", 2),
            qb.H(a[0]),
            qb.H(b[0]),
            qb.H(c[0]),
        )
        ast = slr_to_ast(prog)

        output = pretty_print(ast)

        assert 'a := QReg("a", 2)' in output
        assert 'b := QReg("b", 2)' in output
        assert 'c := QReg("c", 2)' in output

    def test_float_that_looks_like_int(self) -> None:
        """Test float value that equals an integer."""
        expr = LiteralExpr(value=5.0)

        output = format_expression(expr)

        # Should format as integer since it's a whole number
        assert output == "5"



# Copyright 2026 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""AST to QIR (Quantum Intermediate Representation) code generator.

This module transforms AST nodes into QIR using LLVM IR generation.
QIR is an LLVM-based intermediate representation for quantum programs.

Example:
    from pecos.slr.ast import slr_to_ast
    from pecos.slr.ast.codegen import AstToQir

    ast = slr_to_ast(slr_program)
    generator = AstToQir()
    llvm_ir = generator.generate(ast)
"""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import pecos as pc
from pecos.slr.ast.nodes import (
    AllocatorDecl,
    AssignOp,
    BarrierOp,
    BinaryExpr,
    BinaryOp,
    BitExpr,
    BitRef,
    ForStmt,
    GateKind,
    GateOp,
    IfStmt,
    LiteralExpr,
    MeasureOp,
    ParallelBlock,
    PermuteOp,
    PrepareOp,
    RegisterDecl,
    RepeatStmt,
    SlotRef,
    UnaryExpr,
    UnaryOp,
    VarExpr,
    WhileStmt,
)

if TYPE_CHECKING:
    from pecos.slr.ast.nodes import (
        Expression,
        Program,
        Statement,
    )

# Optional LLVM dependency - imported at module level for efficiency
try:
    from pecos_rslib.llvm import ir as llvm_ir

    LLVM_AVAILABLE = True
except ImportError:
    llvm_ir = None  # type: ignore[assignment]
    LLVM_AVAILABLE = False

# Mapping from AST GateKind to QIR gate names
GATE_TO_QIR: dict[GateKind, str] = {
    # Single-qubit Paulis
    GateKind.X: "x",
    GateKind.Y: "y",
    GateKind.Z: "z",
    # Hadamard
    GateKind.H: "h",
    # Phase gates
    GateKind.S: "s",
    GateKind.Sdg: "s__adj",
    GateKind.T: "t",
    GateKind.Tdg: "t__adj",
    # Square root gates - mapped to S variants
    GateKind.SZ: "s",
    GateKind.SZdg: "s__adj",
    # Rotation gates
    GateKind.RX: "rx",
    GateKind.RY: "ry",
    GateKind.RZ: "rz",
    # Two-qubit gates
    GateKind.CX: "cnot",
    GateKind.CZ: "cz",
    GateKind.RZZ: "rzz",
    GateKind.SZZ: "zz",
}

# Gates with rotation parameters
PARAMETERIZED_GATES = {GateKind.RX, GateKind.RY, GateKind.RZ, GateKind.RZZ}

# Two-qubit gates
TWO_QUBIT_GATES = {GateKind.CX, GateKind.CZ, GateKind.RZZ, GateKind.SZZ}


@dataclass
class QirCodeGenContext:
    """Context for QIR code generation."""

    qubit_map: dict[tuple[str, int], int] = field(default_factory=dict)
    qubit_count: int = 0
    creg_map: dict[str, int] = field(default_factory=dict)  # name -> size
    measurement_count: int = 0
    allocator_parents: dict[str, str | None] = field(default_factory=dict)
    allocator_offsets: dict[str, int] = field(default_factory=dict)

    def get_root_allocator(self, name: str) -> str:
        """Get the root allocator for a given allocator name."""
        current = name
        while self.allocator_parents.get(current) is not None:
            current = self.allocator_parents[current]
        return current

    def get_absolute_index(self, allocator: str, index: int) -> int:
        """Get the absolute index in the root allocator."""
        offset = self.allocator_offsets.get(allocator, 0)
        return offset + index

    def get_qubit_index(self, allocator: str, index: int) -> int:
        """Get the global qubit index for an allocator slot.

        For child allocators, translates to root allocator with computed offset.
        """
        # Translate to root allocator and absolute index
        root = self.get_root_allocator(allocator)
        abs_index = self.get_absolute_index(allocator, index)

        key = (root, abs_index)
        if key not in self.qubit_map:
            self.qubit_map[key] = self.qubit_count
            self.qubit_count += 1
        return self.qubit_map[key]


class AstToQir:
    """Transforms AST programs into QIR using recursive descent.

    Generates LLVM IR suitable for QIR-compatible execution environments.

    Usage:
        generator = AstToQir()
        llvm_ir = generator.generate(ast_program)
    """

    def __init__(self) -> None:
        """Initialize the generator."""
        self.context = QirCodeGenContext()
        self._module = None
        self._builder = None
        self._types = None
        self._main_func = None
        self._gate_cache: dict[str, Any] = {}
        self._creg_ptrs: dict[str, Any] = {}
        self._creg_funcs = None

    def generate(self, program: Program) -> str:
        """Generate QIR (LLVM IR) for a program.

        Args:
            program: The AST Program to generate code for.

        Returns:
            QIR as an LLVM IR string.

        Raises:
            ImportError: If LLVM dependencies are not available.
        """
        if not LLVM_AVAILABLE:
            msg = "LLVM dependencies not available. Install with 'pip install pecos[qir]'"
            raise ImportError(msg)

        self.context = QirCodeGenContext()
        self._gate_cache = {}
        self._creg_ptrs = {}

        # Setup LLVM module
        self._module = llvm_ir.Module(name="ast_qir_module")

        # Setup types
        qubit_ty = self._module.context.get_identified_type("Qubit")
        result_ty = self._module.context.get_identified_type("Result")

        self._types = {
            "void": llvm_ir.VoidType(),
            "bool": llvm_ir.IntType(1),
            "int": llvm_ir.IntType(64),
            "double": llvm_ir.DoubleType(),
            "qubit_ptr": qubit_ty.as_pointer(),
            "result_ptr": result_ty.as_pointer(),
            "tag": llvm_ir.IntType(8).as_pointer(),
        }

        # Setup creg helper functions
        self._setup_creg_funcs()

        # Setup measurement function
        self._mz_to_bit = self._declare_function(
            "mz_to_creg_bit",
            self._types["void"],
            [
                self._types["qubit_ptr"],
                self._types["bool"].as_pointer(),
                self._types["int"],
            ],
        )

        # Setup main function
        main_fnty = llvm_ir.FunctionType(self._types["void"], [])
        self._main_func = llvm_ir.Function(self._module, main_fnty, name="main")
        entry_block = self._main_func.append_basic_block(name="entry")
        self._builder = llvm_ir.IRBuilder(entry_block)
        self._builder.comment(
            f"// Generated from AST using: PECOS version {pc.__version__}",
        )

        # Setup operator map
        self._setup_op_map()

        # Process declarations
        self._process_declarations(program)

        # Process body statements
        for stmt in program.body:
            self._process_statement(stmt)

        # Generate results output
        self._generate_results()

        # Return void
        self._builder.ret_void()

        # Return the LLVM IR with attributes
        return self._finalize_module()

    def _setup_creg_funcs(self) -> None:
        """Setup classical register helper functions."""

        self._creg_funcs = {
            "create_creg": self._declare_function(
                "create_creg",
                self._types["bool"].as_pointer(),
                [self._types["int"]],
            ),
            "creg_to_int": self._declare_function(
                "get_int_from_creg",
                self._types["int"],
                [self._types["bool"].as_pointer()],
            ),
            "get_creg_bit": self._declare_function(
                "get_creg_bit",
                self._types["bool"],
                [self._types["bool"].as_pointer(), self._types["int"]],
            ),
            "set_creg_bit": self._declare_function(
                "set_creg_bit",
                self._types["void"],
                [
                    self._types["bool"].as_pointer(),
                    self._types["int"],
                    self._types["bool"],
                ],
            ),
            "set_creg": self._declare_function(
                "set_creg_to_int",
                self._types["void"],
                [self._types["bool"].as_pointer(), self._types["int"]],
            ),
            "int_result": self._declare_function(
                "__quantum__rt__int_record_output",
                self._types["void"],
                [self._types["int"], self._types["tag"]],
            ),
        }

    def _declare_function(self, name: str, ret_ty: Any, arg_tys: list) -> Any:
        """Declare an LLVM function."""
        fnty = llvm_ir.FunctionType(ret_ty, arg_tys)
        return llvm_ir.Function(self._module, fnty, name=name)

    def _setup_op_map(self) -> None:
        """Setup binary operator mapping."""
        self._op_map = {
            BinaryOp.EQ: lambda lhs, rhs: self._builder.icmp_signed("==", lhs, rhs),
            BinaryOp.NE: lambda lhs, rhs: self._builder.icmp_signed("!=", lhs, rhs),
            BinaryOp.LT: lambda lhs, rhs: self._builder.icmp_signed("<", lhs, rhs),
            BinaryOp.GT: lambda lhs, rhs: self._builder.icmp_signed(">", lhs, rhs),
            BinaryOp.LE: lambda lhs, rhs: self._builder.icmp_signed("<=", lhs, rhs),
            BinaryOp.GE: lambda lhs, rhs: self._builder.icmp_signed(">=", lhs, rhs),
            BinaryOp.MUL: self._builder.mul,
            BinaryOp.DIV: self._builder.udiv,
            BinaryOp.XOR: self._builder.xor,
            BinaryOp.AND: self._builder.and_,
            BinaryOp.OR: self._builder.or_,
            BinaryOp.ADD: self._builder.add,
            BinaryOp.SUB: self._builder.sub,
            BinaryOp.RSHIFT: self._builder.lshr,
            BinaryOp.LSHIFT: self._builder.shl,
        }

    def _process_declarations(self, program: Program) -> None:
        """Process declarations to allocate qubits and classical registers."""
        # First pass: collect allocator parent info
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl):
                self.context.allocator_parents[decl.name] = decl.parent

        if program.allocator:
            self.context.allocator_parents[program.allocator.name] = program.allocator.parent

        # Calculate offsets for child allocators
        self._calculate_allocator_offsets(program)

        # Process allocator declarations - only allocate for root allocators
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl):
                if decl.parent is None:
                    for i in range(decl.capacity):
                        self.context.get_qubit_index(decl.name, i)
            elif isinstance(decl, RegisterDecl):
                self.context.creg_map[decl.name] = decl.size
                # Create classical register
                if decl.size < 64:
                    self._creg_ptrs[decl.name] = self._builder.call(
                        self._creg_funcs["create_creg"],
                        [llvm_ir.Constant(self._types["int"], decl.size)],
                        name=decl.name,
                    )

        if program.allocator and program.allocator.parent is None:
            for i in range(program.allocator.capacity):
                self.context.get_qubit_index(program.allocator.name, i)

    def _calculate_allocator_offsets(self, program: Program) -> None:
        """Calculate the offset of each child allocator within its parent."""
        parent_next_offset: dict[str, int] = {}

        # Root allocators have offset 0
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is None:
                self.context.allocator_offsets[decl.name] = 0

        if program.allocator and program.allocator.parent is None:
            self.context.allocator_offsets[program.allocator.name] = 0

        # Process child allocators in declaration order
        for decl in program.declarations:
            if isinstance(decl, AllocatorDecl) and decl.parent is not None:
                parent = decl.parent
                if parent not in parent_next_offset:
                    parent_next_offset[parent] = 0

                parent_offset = self.context.allocator_offsets.get(parent, 0)
                self.context.allocator_offsets[decl.name] = parent_offset + parent_next_offset[parent]
                parent_next_offset[parent] += decl.capacity

    def _process_statement(self, stmt: Statement) -> None:
        """Process a statement using recursive descent."""
        if isinstance(stmt, GateOp):
            self._process_gate(stmt)
        elif isinstance(stmt, MeasureOp):
            self._process_measure(stmt)
        elif isinstance(stmt, PrepareOp):
            self._process_prepare(stmt)
        elif isinstance(stmt, BarrierOp):
            self._process_barrier(stmt)
        elif isinstance(stmt, AssignOp):
            self._process_assign(stmt)
        elif isinstance(stmt, IfStmt):
            self._process_if(stmt)
        elif isinstance(stmt, WhileStmt):
            self._process_while(stmt)
        elif isinstance(stmt, ForStmt):
            self._process_for(stmt)
        elif isinstance(stmt, RepeatStmt):
            self._process_repeat(stmt)
        elif isinstance(stmt, ParallelBlock):
            self._process_parallel(stmt)
        elif isinstance(stmt, PermuteOp):
            self._process_permute(stmt)

    def _process_gate(self, node: GateOp) -> None:
        """Process a gate operation."""
        qir_name = GATE_TO_QIR.get(node.gate)
        if qir_name is None:
            # Skip unsupported gates
            return

        if node.gate in TWO_QUBIT_GATES:
            self._process_two_qubit_gate(node, qir_name)
        else:
            self._process_single_qubit_gate(node, qir_name)

    def _process_single_qubit_gate(self, node: GateOp, qir_name: str) -> None:
        """Process a single-qubit gate."""
        # Get or create gate function
        gate_func = self._get_or_create_gate(
            qir_name,
            has_params=node.gate in PARAMETERIZED_GATES,
            num_qubits=1,
        )

        for target in node.targets:
            qubit_ptr = self._get_qubit_ptr(target)

            args = []
            if node.gate in PARAMETERIZED_GATES and node.params:
                args.extend(llvm_ir.Constant(self._types["double"], float(p)) for p in node.params)
            args.append(qubit_ptr)

            self._builder.call(gate_func, args, name="")

    def _process_two_qubit_gate(self, node: GateOp, qir_name: str) -> None:
        """Process a two-qubit gate."""
        gate_func = self._get_or_create_gate(
            qir_name,
            has_params=node.gate in PARAMETERIZED_GATES,
            num_qubits=2,
        )

        if len(node.targets) >= 2:
            q0_ptr = self._get_qubit_ptr(node.targets[0])
            q1_ptr = self._get_qubit_ptr(node.targets[1])

            args = []
            if node.gate in PARAMETERIZED_GATES and node.params:
                args.extend(llvm_ir.Constant(self._types["double"], float(p)) for p in node.params)
            args.extend([q0_ptr, q1_ptr])

            self._builder.call(gate_func, args, name="")

    def _get_or_create_gate(
        self,
        qir_name: str,
        *,
        has_params: bool,
        num_qubits: int,
    ) -> Any:
        """Get or create a QIR gate function declaration."""
        cache_key = f"{qir_name}_{has_params}_{num_qubits}"
        if cache_key in self._gate_cache:
            return self._gate_cache[cache_key]

        # Build argument types
        arg_tys = []
        if has_params:
            arg_tys.append(self._types["double"])
        arg_tys.extend([self._types["qubit_ptr"]] * num_qubits)

        # Build mangled name
        suffix = "__body" if "adj" not in qir_name else ""
        mangled_name = f"__quantum__qis__{qir_name}{suffix}"

        fnty = llvm_ir.FunctionType(self._types["void"], arg_tys)
        gate_func = llvm_ir.Function(self._module, fnty, name=mangled_name)

        self._gate_cache[cache_key] = gate_func
        return gate_func

    def _get_qubit_ptr(self, target: SlotRef) -> Any:
        """Get a qubit pointer for a target."""
        qubit_index = self.context.get_qubit_index(target.allocator, target.index)
        return llvm_ir.Constant(self._types["int"], qubit_index).inttoptr(
            self._types["qubit_ptr"],
        )

    def _process_measure(self, node: MeasureOp) -> None:
        """Process a measurement operation."""
        for i, target in enumerate(node.targets):
            self.context.measurement_count += 1
            qubit_ptr = self._get_qubit_ptr(target)

            if i < len(node.results):
                result = node.results[i]
                if result.register in self._creg_ptrs:
                    creg_ptr = self._creg_ptrs[result.register]
                    bit_index = llvm_ir.Constant(self._types["int"], result.index)
                    self._builder.call(
                        self._mz_to_bit,
                        [qubit_ptr, creg_ptr, bit_index],
                        name="",
                    )

    def _process_prepare(self, node: PrepareOp) -> None:
        """Process a prepare/reset operation."""
        if node.slots is None:
            return

        reset_func = self._get_or_create_gate("reset", has_params=False, num_qubits=1)

        for slot in node.slots:
            qubit_ptr = self._get_qubit_ptr(
                SlotRef(allocator=node.allocator, index=slot),
            )
            self._builder.call(reset_func, [qubit_ptr], name="")

    def _process_barrier(self, node: BarrierOp) -> None:
        """Process a barrier operation."""
        # Collect all qubits involved
        qubits = []
        if node.allocators:
            for alloc in node.allocators:
                qubits.extend(
                    self._get_qubit_ptr(SlotRef(allocator=key[0], index=key[1]))
                    for key in self.context.qubit_map
                    if key[0] == alloc
                )

        if not qubits:
            return

        # Create barrier function if needed
        barrier_name = f"__quantum__qis__barrier{len(qubits)}__body"
        fnty = llvm_ir.FunctionType(
            self._types["void"],
            [self._types["qubit_ptr"]] * len(qubits),
        )
        barrier_func = llvm_ir.Function(self._module, fnty, name=barrier_name)
        self._builder.call(barrier_func, qubits, name="")

    def _process_assign(self, node: AssignOp) -> None:
        """Process an assignment operation."""
        if isinstance(node.target, BitRef):
            reg_name = node.target.register
            if reg_name not in self._creg_ptrs:
                return

            creg_ptr = self._creg_ptrs[reg_name]
            bit_index = llvm_ir.Constant(self._types["int"], node.target.index)

            # Evaluate RHS
            rhs = self._eval_expression(node.value)

            self._builder.call(
                self._creg_funcs["set_creg_bit"],
                [creg_ptr, bit_index, rhs],
                name="",
            )

    def _eval_expression(self, expr: Expression) -> Any:
        """Evaluate an expression to an LLVM value."""
        if isinstance(expr, LiteralExpr):
            if isinstance(expr.value, bool):
                return llvm_ir.Constant(self._types["bool"], 1 if expr.value else 0)
            if isinstance(expr.value, int):
                return llvm_ir.Constant(self._types["int"], expr.value)
            if isinstance(expr.value, float):
                return llvm_ir.Constant(self._types["double"], expr.value)
            return llvm_ir.Constant(self._types["int"], expr.value)

        if isinstance(expr, BitExpr):
            reg_name = expr.ref.register
            if reg_name not in self._creg_ptrs:
                return llvm_ir.Constant(self._types["bool"], 0)
            creg_ptr = self._creg_ptrs[reg_name]
            bit_index = llvm_ir.Constant(self._types["int"], expr.ref.index)
            return self._builder.call(
                self._creg_funcs["get_creg_bit"],
                [creg_ptr, bit_index],
                name="",
            )

        if isinstance(expr, VarExpr):
            # Variable lookup - for now just return 0
            return llvm_ir.Constant(self._types["int"], 0)

        if isinstance(expr, BinaryExpr):
            left = self._eval_expression(expr.left)
            right = self._eval_expression(expr.right)
            if expr.op in self._op_map:
                return self._op_map[expr.op](left, right)
            return left

        if isinstance(expr, UnaryExpr):
            operand = self._eval_expression(expr.operand)
            if expr.op == UnaryOp.NEG:
                return self._builder.neg(operand)
            if expr.op == UnaryOp.NOT:
                return self._builder.not_(operand)
            return operand

        return llvm_ir.Constant(self._types["int"], 0)

    def _process_if(self, node: IfStmt) -> None:
        """Process an if statement."""
        pred = self._eval_expression(node.condition)

        if node.else_body:
            with self._builder.if_else(pred) as (then, otherwise):
                with then:
                    for stmt in node.then_body:
                        self._process_statement(stmt)
                with otherwise:
                    for stmt in node.else_body:
                        self._process_statement(stmt)
        else:
            with self._builder.if_then(pred):
                for stmt in node.then_body:
                    self._process_statement(stmt)

    def _process_while(self, node: WhileStmt) -> None:
        """Process a while loop."""
        # QIR supports loops through LLVM branch instructions
        # For simplicity, we process the body once (approximation)
        for stmt in node.body:
            self._process_statement(stmt)

    def _process_for(self, node: ForStmt) -> None:
        """Process a for loop by unrolling."""
        if isinstance(node.start, int) and isinstance(node.stop, int):
            step = node.step if isinstance(node.step, int) else 1
            for _ in range(node.start, node.stop, step):
                for stmt in node.body:
                    self._process_statement(stmt)

    def _process_repeat(self, node: RepeatStmt) -> None:
        """Process a repeat loop by unrolling."""
        if isinstance(node.count, int):
            for _ in range(node.count):
                for stmt in node.body:
                    self._process_statement(stmt)

    def _process_parallel(self, node: ParallelBlock) -> None:
        """Process a parallel block."""
        for stmt in node.body:
            self._process_statement(stmt)

    def _process_permute(self, node: PermuteOp) -> None:
        """Process a permutation operation.

        Updates the internal allocator mapping to swap qubit references.
        QIR doesn't have a permute instruction, so this just updates
        how we map allocator names to qubit indices.
        """
        # Swap the allocator offsets
        for src, tgt in zip(node.sources, node.targets, strict=False):
            # Get current offsets
            src_offset = self.context.allocator_offsets.get(src, 0)
            tgt_offset = self.context.allocator_offsets.get(tgt, 0)
            # Swap them
            self.context.allocator_offsets[src] = tgt_offset
            self.context.allocator_offsets[tgt] = src_offset

    def _generate_results(self) -> None:
        """Generate result output calls."""
        for reg_name, creg_ptr in self._creg_ptrs.items():
            # Create tag for the register name
            reg_name_bytes = bytearray(reg_name.encode("utf-8"))
            tag_type = llvm_ir.ArrayType(llvm_ir.IntType(8), len(reg_name))
            reg_tag = llvm_ir.GlobalVariable(self._module, tag_type, reg_name)
            reg_tag.initializer = llvm_ir.Constant(tag_type, reg_name_bytes)
            reg_tag.global_constant = True
            reg_tag.linkage = "private"

            # Convert creg to int and output
            c_int = self._builder.call(
                self._creg_funcs["creg_to_int"],
                [creg_ptr],
                name="",
            )
            reg_tag_gep = reg_tag.gep(
                (
                    llvm_ir.Constant(llvm_ir.IntType(32), 0),
                    llvm_ir.Constant(llvm_ir.IntType(32), 0),
                ),
            )
            self._builder.call(
                self._creg_funcs["int_result"],
                [c_int, reg_tag_gep],
                name="",
            )

    def _finalize_module(self) -> str:
        """Finalize the module and return LLVM IR with attributes."""
        ll_text = self._fix_internal_consts(str(self._module))
        mod_w_attr = ll_text.replace("@main()", "@main() #0")

        mod_w_attr += '\nattributes #0 = { "entry_point"'
        mod_w_attr += ' "qir_profiles"="custom"'
        mod_w_attr += f' "required_num_qubits"="{self.context.qubit_count}"'
        mod_w_attr += f' "required_num_results"="{self.context.measurement_count}" }}'
        return mod_w_attr

    def _fix_internal_consts(self, llvm_ir: str) -> str:
        """Fix internal constants in LLVM IR."""
        return re.sub('([@%])"([^"]+)"', r"\1\2", llvm_ir)


def ast_to_qir(program: Program) -> str:
    """Convert an AST Program to QIR (LLVM IR).

    Convenience function for simple code generation.

    Args:
        program: The AST Program to convert.

    Returns:
        QIR as an LLVM IR string.
    """
    generator = AstToQir()
    return generator.generate(program)



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Internal utilities for PECOS.

This package contains internal implementation details that are not part of
the public API. These modules should not be imported directly by users.

**WARNING**: The contents of this package are subject to change without notice.
Do not rely on any APIs defined here.

Internal modules (re-exports from misc for backwards compatibility):
    - commute: Commutation relation utilities
    - stabilizer_funcs: Stabilizer manipulation functions
    - threshold_curve: Threshold curve fitting
    - symbol_library: Gate symbol library definitions
    - std_output: Standard output formatting
    - gate_groups: Gate classification utilities
"""

# Re-export internal modules from misc for backwards compatibility
# These are internal APIs and subject to change
from pecos.misc import commute
from pecos.misc.gate_groups import (
    error_one_paulis_collection,
    error_two_paulis_collection,
    one_qubits,
    two_qubits,
)
from pecos.misc.stabilizer_funcs import (
    circ2set,
    find_stab,
    op_commutes,
    remove_stab,
)
from pecos.misc.std_output import StdOutput
from pecos.misc.symbol_library import SymbolLibrary
from pecos.misc.threshold_curve import func as threshold_func
from pecos.misc.threshold_curve import threshold_fit

__all__ = [
    # Output
    "StdOutput",
    # Symbol library
    "SymbolLibrary",
    # Stabilizer functions
    "circ2set",
    # Commute module
    "commute",
    # Gate groups
    "error_one_paulis_collection",
    "error_two_paulis_collection",
    "find_stab",
    "one_qubits",
    "op_commutes",
    "remove_stab",
    # Threshold fitting
    "threshold_fit",
    "threshold_func",
    "two_qubits",
]



"""Base classes for surface code layouts."""

from enum import Enum
from typing import Protocol

from pecos.slr.qeclib.surface.visualization.visualization_base import VisualizationData


class LatticeType(Enum):
    """Lattices that the patches of the surface code can be constructed from.

    References:
        1. Jonas Anderson, "Fault-tolerance in two-dimensional topological systems" by
           <https://digitalrepository.unm.edu/phyc_etds/4/>
    """

    SQUARE = (4, 4, 4, 4)
    RHOMBITRIHEXAGONAL = (3, 4, 6, 4)
    TRIHEXAGONAL = (3, 6, 3, 6)


class Layout(Protocol):
    """Protocol for different layout strategies."""

    def get_stabilizers_gens(
        self,
        dx: int,
        dz: int,
    ) -> list[tuple[str, tuple[int, ...]]]:
        """Get stabilizer generators for the layout."""
        ...

    def get_data_positions(self, dx: int, dz: int) -> list[tuple[int, int]]:
        """Get positions of data qubits in the layout."""
        ...

    def validate_dimensions(self, dx: int, dz: int) -> None:
        """Validate the layout dimensions."""
        ...

    def get_visualization_elements(
        self,
        dx: int,
        dz: int,
        stab_gens: list[tuple[str, tuple[int, ...]]],
    ) -> VisualizationData:
        """Get visualization elements for the layout."""
        ...



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Analysis tools for quantum error correction.

This package provides comprehensive analysis tools for quantum error correction,
including threshold estimation, fault tolerance verification, and stabilizer
code analysis.

Submodules:
    threshold: Threshold estimation and code capacity analysis
    fault_tolerance: Fault tolerance checking and verification
    stabilizers: Stabilizer code verification and distance analysis

Example:
    >>> from pecos.analysis import threshold, fault_tolerance
    >>> from pecos.analysis.stabilizers import VerifyStabilizers
    >>>
    >>> # Threshold analysis
    >>> result = threshold.code_capacity(
    ...     qecc_class=Surface4444,
    ...     error_gen=XModel,
    ...     decoder_class=MWPM2D,
    ...     ps=[0.01, 0.05, 0.10],
    ...     ds=[3, 5, 7],
    ...     runs=1000,
    ... )
    >>>
    >>> # Fault tolerance verification
    >>> passed, weight = fault_tolerance.t_errors_check(qecc)
"""

# Re-export from tools for backwards compatibility
from pecos.tools import fault_tolerance_checks as fault_tolerance
from pecos.tools import pseudo_threshold_tools as pseudo_threshold
from pecos.tools import threshold_tools as threshold
from pecos.tools.stabilizer_verification import VerifyStabilizers
from pecos.tools.threshold_tools import (
    codecapacity_logical_rate,
    codecapacity_logical_rate2,
    codecapacity_logical_rate3,
    threshold_code_capacity,
)

__all__ = [
    # Classes
    "VerifyStabilizers",
    # Threshold functions
    "codecapacity_logical_rate",
    "codecapacity_logical_rate2",
    "codecapacity_logical_rate3",
    # Submodules
    "fault_tolerance",
    "pseudo_threshold",
    "threshold",
    "threshold_code_capacity",
]



"""Test register-wide operations that should generate loops."""

from pecos.slr import CReg, Main, QReg, SlrConverter
from pecos.slr.qeclib import qubit
from pecos.slr.qeclib.qubit.measures import Measure


def test_hadamard_on_register() -> None:
    """Test that H(q) generates a loop when q is a register."""
    prog = Main(
        q := QReg("q", 4),
        # Apply Hadamard to entire register
        qubit.H(q),
        # Measure all qubits
        Measure(q) > CReg("c", 4),
    )

    guppy_code = SlrConverter(prog).guppy()

    # Should generate a loop to apply H to each qubit
    assert "for" in guppy_code or "quantum.h(q[0])" in guppy_code

    # Check that H is applied (either in a loop or expanded)
    if "for" in guppy_code:
        # Loop form
        assert "quantum.h(q[i])" in guppy_code
    else:
        # Expanded form
        h_count = guppy_code.count("quantum.h")
        assert h_count >= 4, f"Expected at least 4 H gates, got {h_count}"


def test_multiple_gates_on_register() -> None:
    """Test multiple single-qubit gates on registers."""
    prog = Main(
        q := QReg("q", 3),
        # Apply multiple gates to entire register
        qubit.H(q),
        qubit.X(q),
        qubit.Z(q),
        # Measure all
        Measure(q) > CReg("c", 3),
    )

    guppy_code = SlrConverter(prog).guppy()

    # Check that all gates are applied (either in loops or expanded)
    if "for" in guppy_code:
        # Loop form
        assert "quantum.h(q[i])" in guppy_code
        assert "quantum.x(q[i])" in guppy_code
        assert "quantum.z(q[i])" in guppy_code
    else:
        # Expanded form
        assert guppy_code.count("quantum.h") >= 3
        assert guppy_code.count("quantum.x") >= 3
        assert guppy_code.count("quantum.z") >= 3


def test_mixed_register_and_element_ops() -> None:
    """Test mixing register-wide and element-specific operations."""
    prog = Main(
        q := QReg("q", 4),
        # Apply H to entire register
        qubit.H(q),
        # Apply X to specific elements
        qubit.X(q[0]),
        qubit.X(q[2]),
        # Apply Z to entire register again
        qubit.Z(q),
        Measure(q) > CReg("c", 4),
    )

    guppy_code = SlrConverter(prog).guppy()

    # Should have H and Z applied to all qubits (either in loops or expanded)
    if "for" in guppy_code:
        # Loop form - count loops
        assert "quantum.h(q[i])" in guppy_code
        assert "quantum.z(q[i])" in guppy_code
    else:
        # Expanded form
        assert guppy_code.count("quantum.h") >= 4
        assert guppy_code.count("quantum.z") >= 4

    # Should have X applied to specific qubits (always individual)
    assert "quantum.x(q[0])" in guppy_code
    assert "quantum.x(q[2])" in guppy_code



# Copyright 2024 The PECOS Developers
# Licensed under the Apache License, Version 2.0

"""Tests for pecos.qec.surface layout functions."""

import pytest
from pecos.qec.surface.layouts import (
    StabilizerSupport,
    compute_x_stabilizer_supports,
    compute_z_stabilizer_supports,
    generate_nonrotated_surface_layout,
    generate_surface_layout,
    get_boundary_stabilizer_indices,
    get_boundary_stabilizers,
    get_bulk_stabilizer_indices,
    get_bulk_stabilizers,
    get_stabilizer_counts,
)


class TestGenerateNonrotatedSurfaceLayout:
    """Tests for non-rotated surface code layout generation."""

    def test_d3_data_qubit_count(self) -> None:
        """Distance-3 non-rotated code should have 13 data qubits."""
        data, _ancilla, _polygons = generate_nonrotated_surface_layout(3, 3)
        assert len(data) == 13

    def test_d5_data_qubit_count(self) -> None:
        """Distance-5 non-rotated code should have 41 data qubits."""
        data, _ancilla, _polygons = generate_nonrotated_surface_layout(5, 5)
        assert len(data) == 41

    def test_d3_total_positions(self) -> None:
        """Check total positions for d=3."""
        data, ancilla, _polygons = generate_nonrotated_surface_layout(3, 3)
        # 5x5 lattice = 25 total positions
        assert len(data) + len(ancilla) == 25

    def test_asymmetric_layout(self) -> None:
        """Test asymmetric width/height."""
        data, ancilla, _polygons = generate_nonrotated_surface_layout(3, 5)
        # 5x9 lattice
        assert len(data) + len(ancilla) == 45

    def test_data_positions_are_unique(self) -> None:
        """All data positions should be unique."""
        data, _, _ = generate_nonrotated_surface_layout(5, 5)
        assert len(data) == len(set(data))

    def test_ancilla_positions_are_unique(self) -> None:
        """All ancilla positions should be unique."""
        _, ancilla, _ = generate_nonrotated_surface_layout(5, 5)
        assert len(ancilla) == len(set(ancilla))

    def test_no_overlap_data_ancilla(self) -> None:
        """Data and ancilla positions should not overlap."""
        data, ancilla, _ = generate_nonrotated_surface_layout(5, 5)
        data_set = set(data)
        ancilla_set = set(ancilla)
        assert data_set.isdisjoint(ancilla_set)


class TestGenerateSurfaceLayout:
    """Tests for rotated surface code layout generation (default)."""

    def test_d3_data_qubit_count(self) -> None:
        """Distance-3 rotated code should have 9 data qubits."""
        data, _ancilla = generate_surface_layout(3, 3)
        assert len(data) == 9

    def test_d5_data_qubit_count(self) -> None:
        """Distance-5 rotated code should have 25 data qubits."""
        data, _ancilla = generate_surface_layout(5, 5)
        assert len(data) == 25

    def test_d3_ancilla_count(self) -> None:
        """Distance-3 rotated code should have 8 ancilla qubits."""
        _data, ancilla = generate_surface_layout(3, 3)
        assert len(ancilla) == 8

    def test_data_positions_are_unique(self) -> None:
        """All data positions should be unique."""
        data, _ = generate_surface_layout(5, 5)
        assert len(data) == len(set(data))

    def test_ancilla_positions_are_unique(self) -> None:
        """All ancilla positions should be unique."""
        _, ancilla = generate_surface_layout(5, 5)
        assert len(ancilla) == len(set(ancilla))

    def test_no_overlap_data_ancilla(self) -> None:
        """Data and ancilla positions should not overlap."""
        data, ancilla = generate_surface_layout(5, 5)
        data_set = set(data)
        ancilla_set = set(ancilla)
        assert data_set.isdisjoint(ancilla_set)

    def test_data_at_odd_odd_positions(self) -> None:
        """In interior, data qubits should be at odd-odd positions."""
        data, _ = generate_surface_layout(5, 5)
        for x, y in data:
            # Interior data qubits are at odd-odd positions
            if 0 < x < 10 and 0 < y < 10:
                assert x % 2 == 1
                assert y % 2 == 1


class TestComputeStabilizerSupports:
    """Tests for stabilizer support computation."""

    def test_x_stabilizer_count_d3(self) -> None:
        """Distance-3 should have 4 X stabilizers."""
        stabs = compute_x_stabilizer_supports(3)
        assert len(stabs) == 4

    def test_z_stabilizer_count_d3(self) -> None:
        """Distance-3 should have 4 Z stabilizers."""
        stabs = compute_z_stabilizer_supports(3)
        assert len(stabs) == 4

    def test_x_stabilizer_count_d5(self) -> None:
        """Distance-5 should have 12 X stabilizers."""
        stabs = compute_x_stabilizer_supports(5)
        assert len(stabs) == 12

    def test_z_stabilizer_count_d5(self) -> None:
        """Distance-5 should have 12 Z stabilizers."""
        stabs = compute_z_stabilizer_supports(5)
        assert len(stabs) == 12

    def test_stabilizer_indices_are_sequential(self) -> None:
        """Stabilizer indices should be 0 to n-1."""
        stabs = compute_x_stabilizer_supports(5)
        indices = sorted(s.index for s in stabs)
        assert indices == list(range(len(stabs)))

    def test_bulk_stabilizers_have_weight_4(self) -> None:
        """Non-boundary stabilizers should have weight 4."""
        stabs = compute_x_stabilizer_supports(5)
        for s in stabs:
            if not s.is_boundary:
                assert s.weight == 4

    def test_boundary_stabilizers_have_weight_2(self) -> None:
        """Boundary stabilizers should have weight 2."""
        stabs = compute_x_stabilizer_supports(5)
        for s in stabs:
            if s.is_boundary:
                assert s.weight == 2

    def test_stabilizer_support_is_dataclass(self) -> None:
        """StabilizerSupport should be a frozen dataclass."""
        stabs = compute_x_stabilizer_supports(3)
        s = stabs[0]
        assert isinstance(s, StabilizerSupport)
        # Frozen dataclass should be hashable
        assert hash(s) is not None


class TestStabilizerCategories:
    """Tests for bulk/boundary stabilizer categorization."""

    def test_stabilizer_counts_d3(self) -> None:
        """Test stabilizer counts for d=3."""
        total, n_bulk, n_boundary = get_stabilizer_counts(3)
        assert total == 4
        assert n_bulk == 2
        assert n_boundary == 2

    def test_stabilizer_counts_d5(self) -> None:
        """Test stabilizer counts for d=5."""
        total, n_bulk, n_boundary = get_stabilizer_counts(5)
        assert total == 12
        assert n_bulk == 8
        assert n_boundary == 4

    def test_stabilizer_counts_d7(self) -> None:
        """Test stabilizer counts for d=7."""
        total, n_bulk, n_boundary = get_stabilizer_counts(7)
        assert total == 24
        assert n_bulk == 18
        assert n_boundary == 6

    def test_bulk_indices_d3(self) -> None:
        """Bulk indices for d=3 should be [1, 2]."""
        indices = get_bulk_stabilizer_indices(3)
        assert indices == [1, 2]

    def test_boundary_indices_d3(self) -> None:
        """Boundary indices for d=3 should be [0, 3]."""
        indices = get_boundary_stabilizer_indices(3)
        assert indices == [0, 3]

    def test_bulk_indices_d5(self) -> None:
        """Bulk indices for d=5."""
        indices = get_bulk_stabilizer_indices(5)
        assert len(indices) == 8
        assert indices == list(range(2, 10))

    def test_boundary_indices_d5(self) -> None:
        """Boundary indices for d=5."""
        indices = get_boundary_stabilizer_indices(5)
        assert len(indices) == 4
        assert indices == [0, 1, 10, 11]

    def test_bulk_plus_boundary_equals_total(self) -> None:
        """Bulk + boundary indices should cover all stabilizers."""
        for d in [3, 5, 7]:
            bulk = get_bulk_stabilizer_indices(d)
            boundary = get_boundary_stabilizer_indices(d)
            total, _, _ = get_stabilizer_counts(d)
            assert len(bulk) + len(boundary) == total
            assert set(bulk).isdisjoint(set(boundary))

    def test_get_bulk_stabilizers_x(self) -> None:
        """Get bulk X stabilizers."""
        bulk = get_bulk_stabilizers(5, "X")
        assert len(bulk) == 8
        for s in bulk:
            assert not s.is_boundary
            assert s.weight == 4

    def test_get_boundary_stabilizers_x(self) -> None:
        """Get boundary X stabilizers."""
        boundary = get_boundary_stabilizers(5, "X")
        assert len(boundary) == 4
        for s in boundary:
            assert s.is_boundary
            assert s.weight == 2

    def test_get_bulk_stabilizers_z(self) -> None:
        """Get bulk Z stabilizers."""
        bulk = get_bulk_stabilizers(5, "Z")
        assert len(bulk) == 8
        for s in bulk:
            assert not s.is_boundary
            assert s.weight == 4

    def test_get_boundary_stabilizers_z(self) -> None:
        """Get boundary Z stabilizers."""
        boundary = get_boundary_stabilizers(5, "Z")
        assert len(boundary) == 4
        for s in boundary:
            assert s.is_boundary
            assert s.weight == 2



"""Test and document default values for QASM simulations using sim() API."""


class TestQasmSimDefaults:
    """Test and document default values for all QASM simulation settings."""

    def test_builder_defaults(self) -> None:
        """Test and document defaults when using qasm_engine builder."""
        from pecos import Qasm, qasm_engine

        qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        creg c[2];
        h q[0];
        cx q[0], q[1];
        measure q -> c;
        """

        # Build with all defaults
        sim = qasm_engine().program(Qasm.from_string(qasm)).to_sim().build()

        # Based on Rust code, the defaults are:
        # - seed: None (non-deterministic)
        # - workers: 1 (single thread)
        # - noise_model: no noise (don't call .noise())
        # - quantum_engine: SparseStabilizer
        # - bit_format: BigInt (integers, not binary strings)

        # Run to verify it works
        results = sim.run(100)
        results_dict = results.to_dict()
        assert len(results_dict["c"]) == 100

    def test_run_direct_defaults(self) -> None:
        """Test and document defaults when using engine run directly."""
        from pecos import Qasm, qasm_engine

        qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[1];
        creg c[1];
        x q[0];
        measure q[0] -> c[0];
        """

        # Run with minimal parameters using new API
        results = qasm_engine().program(Qasm.from_string(qasm)).to_sim().run(10)
        results_dict = results.to_dict()

        # Defaults for direct run:
        # - noise_model: None (no noise)
        # - engine: auto-selected based on circuit
        # - workers: defaults to 1
        # - seed: None (non-deterministic)

        assert all(val == 1 for val in results_dict["c"])

    def test_noise_model_defaults(self) -> None:
        """Test and document default parameters for noise models."""
        from pecos import (
            GeneralNoiseModelBuilder,
            biased_depolarizing_noise,
            depolarizing_noise,
        )

        # Test default values for noise models using builder pattern
        # Note: depolarizing_noise() builder requires explicit probability
        depolarizing_noise().with_p1_probability(0.001)
        # Can't directly assert on builder properties

        # General noise model has defaults that can be overridden
        GeneralNoiseModelBuilder()
        # Default values are set when building

        (biased_depolarizing_noise().with_p1_probability(0.001).with_p2_probability(0.001).with_prep_probability(0.001))
        # Builder pattern requires explicit values

    def test_builder_defaults_new_api(self) -> None:
        """Test and document defaults when using new unified API."""
        from pecos import Qasm, qasm_engine

        # Minimal setup - only required field
        qasm = """
            OPENQASM 2.0;
            include "qelib1.inc";
            qreg q[1];
            creg c[1];
            x q[0];
            measure q[0] -> c[0];
            """

        sim = qasm_engine().program(Qasm.from_string(qasm)).to_sim().build()
        results = sim.run(10)
        results_dict = results.to_dict()

        # Defaults for new API:
        # - seed: None (not set)
        # - workers: 1 (default)
        # - noise: no noise (ideal simulation)
        # - quantum_engine: SparseStabilizer (default)
        # - binary_string_format: False (integers)

        assert all(val == 1 for val in results_dict["c"])

    def test_no_noise_means_ideal(self) -> None:
        """Test that omitting noise results in ideal (deterministic) simulation."""
        from pecos import Qasm, qasm_engine

        qasm = """
        OPENQASM 2.0;
        include "qelib1.inc";
        qreg q[2];
        creg c[2];
        x q[0];
        x q[1];
        measure q -> c;
        """

        # Build without noise specification
        sim1 = qasm_engine().program(Qasm.from_string(qasm)).to_sim().build()

        # Both should produce identical deterministic results
        results1 = sim1.run(100)
        results1_dict = results1.to_dict()

        # Should always measure |11> = 3
        assert all(val == 3 for val in results1_dict["c"])

    def test_default_summary(self) -> None:
        """Document all defaults in one place."""
        # Default values summary:
        #
        # QasmEngine defaults:
        # - seed: None (non-deterministic)
        # - workers: 1 (single thread)
        # - noise_model: no noise (ideal simulation)
        # - quantum_engine: SparseStabilizer
        # - bit_format: BigInt (integers, not binary strings)
        #
        # Noise model builders:
        # - depolarizing_noise(): requires explicit .with_p1_probability()
        # - biased_depolarizing_noise(): requires probability settings
        # - GeneralNoiseModelBuilder(): has internal defaults
        #
        # New unified API defaults:
        # - All optional fields use builder defaults when not specified
        # - noise: no noise (ideal simulation) when omitted

        # This test just documents the defaults
        assert True



# Copyright 2025 The PECOS Developers
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
# the License.You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

"""Tests for SlrConverter with ParallelOptimizer integration."""

from pecos.slr import Block, CReg, Main, Parallel, QReg, SlrConverter
from pecos.slr.qeclib import qubit as qb


def test_slr_converter_without_optimization() -> None:
    """Test SlrConverter without optimization."""
    prog = Main(
        q := QReg("q", 4),
        Parallel(
            qb.H(q[0]),
            qb.X(q[1]),
            qb.H(q[2]),
            qb.X(q[3]),
        ),
    )

    # Explicitly disable optimization
    qasm = SlrConverter(prog, optimize_parallel=False).qasm()

    # Operations should appear in original order (with parallel comments)
    ops = [
        line.strip()
        for line in qasm.split("\n")
        if line.strip() and not line.startswith(("OPENQASM", "include", "qreg", "creg", "//"))
    ]
    assert ops == ["h q[0];", "x q[1];", "h q[2];", "x q[3];"]


def test_slr_converter_with_optimization() -> None:
    """Test SlrConverter with optimization (default behavior)."""
    prog = Main(
        q := QReg("q", 4),
        Parallel(
            qb.H(q[0]),
            qb.X(q[1]),
            qb.H(q[2]),
            qb.X(q[3]),
        ),
    )

    # Default: optimization enabled
    qasm = SlrConverter(prog).qasm()

    # Operations should be reordered by type (H gates first, then X gates)
    # AST codegen adds parallel comments
    ops = [
        line.strip()
        for line in qasm.split("\n")
        if line.strip() and not line.startswith(("OPENQASM", "include", "qreg", "creg", "//"))
    ]
    assert ops == ["h q[0];", "h q[2];", "x q[1];", "x q[3];"]


def test_slr_converter_optimization_preserves_semantics() -> None:
    """Test that optimization preserves the semantic meaning of the circuit."""
    prog = Main(
        q := QReg("q", 4),
        c := CReg("m", 4),
        Parallel(
            Block(
                qb.H(q[0]),
                qb.CX(q[0], q[1]),
            ),
            Block(
                qb.H(q[2]),
                qb.CX(q[2], q[3]),
            ),
        ),
        qb.Measure(q) > c,
    )

    # Generate both versions
    qasm_unopt = SlrConverter(prog, optimize_parallel=False).qasm()
    qasm_opt = SlrConverter(prog).qasm()

    # Both should have the same gates
    # AST codegen uses individual measure statements
    for gate in [
        "h q[0]",
        "h q[2]",
        "cx q[0], q[1]",
        "cx q[2], q[3]",
        "measure q[0] -> m[0]",
    ]:
        assert gate in qasm_unopt
        assert gate in qasm_opt

    # Optimized version should have H gates before CX gates
    assert qasm_opt.index("h q[0]") < qasm_opt.index("cx q[0], q[1]")
    assert qasm_opt.index("h q[2]") < qasm_opt.index("cx q[2], q[3]")
    assert qasm_opt.index("h q[0]") < qasm_opt.index(
        "cx q[2], q[3]",
    )  # Cross-pair ordering


def test_slr_converter_no_parallel_blocks() -> None:
    """Test that programs without Parallel blocks are unaffected by optimization setting."""
    prog = Main(
        q := QReg("q", 4),
        qb.H(q[0]),
        qb.CX(q[0], q[1]),
        qb.H(q[2]),
        qb.CX(q[2], q[3]),
    )

    # Both should produce identical output since there are no Parallel blocks
    qasm_default = SlrConverter(prog).qasm()
    qasm_no_opt = SlrConverter(prog, optimize_parallel=False).qasm()

    assert qasm_default == qasm_no_opt
